"""
Manage configuration files
"""

from .. import errors, utils
from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class SetJob(JobBase):
    """Change or show option in configuration file"""

    name = 'set'
    label = 'Set'
    hidden = True
    cache_id = None  # Don't cache output

    def initialize(self, *, config, option=None, value='', reset=None):
        """
        Set and display option(s)

        :param config: :class:`~.configfiles.ConfigFiles` instance
        :param str option: "."-delimited path to option in `config` or `None`
        :param value: New value for `option` or any falsy value to display the
            current value
        :param bool reset: Whether to reset `option` to default value and ignore
            `value`

        If only `config` is given, display all options and values.

        If `option` is given, display only its value.

        If `option` and `value` is given, set `option` to `value` display the
        result.
        """
        if value and reset:
            raise RuntimeError('Arguments "value" and "reset" are mutually exclusive.')
        elif value and not option:
            raise RuntimeError('Argument "value" needs argument "option".')
        else:
            self._config = config
            self._option = option
            self._value = value
            self._reset = reset

    def execute(self):
        try:
            if self._reset:
                self._reset_mode()
            elif self._value:
                self._set_mode()
            else:
                self._display_mode()
        except errors.ConfigError as e:
            self.error(e)
        finally:
            self.finish()

    def _reset_mode(self):
        if self._option:
            self._config.reset(self._option)
            self._write(self._option)
        else:
            for o in self._config.paths:
                self._config.reset(o)
                self._write(o)

    def _set_mode(self):
        self._config[self._option] = self._value
        self._write(self._option)

    def _display_mode(self):
        if self._option:
            self._display_option(self._option)
        else:
            for o in self._config.paths:
                self._display_option(o)

    def _write(self, option):
        self._config.write(option)
        self._display_option(option)

    def _display_option(self, option):
        if utils.is_sequence(self._config[option]):
            values = '\n  '.join(str(v) for v in self._config[option])
            if values:
                self.send(f'{option} =\n  ' + values)
            else:
                self.send(f'{option} =')
        else:
            self.send(f'{option} = {self._config[option]}')



"""
Concrete :class:`~.base.TrackerConfigBase`, :class:`~.base.TrackerJobsBase`
and :class:`~.base.TrackerBase` subclasses for BHD
"""

from .config import BhdTrackerConfig
from .jobs import BhdTrackerJobs
from .tracker import BhdTracker



"""
Share generated metadata
"""

import asyncio

from .. import errors, trackers
from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class SubmitJob(JobBase):
    """
    Submit torrent file and other metadata to tracker

    This job adds the following signals to the :attr:`~.JobBase.signal`
    attribute:

        ``logging_in``
            Emitted when attempting to start a user session. Registered
            callbacks get no arguments.

        ``logged_in``
            Emitted when login attempt ended. Registered callbacks get no
            arguments.

        ``uploading``
            Emitted when attempting to upload metadata. Registered callbacks get
            no arguments.

        ``uploaded``
            Emitted when upload attempt ended. Registered callbacks get no
            arguments.

        ``logging_out``
            Emitted when attempting to end a user session. Registered callbacks
            get no arguments.

        ``logged_out``
            Emitted when logout attempt ended. Registered callbacks get no
            arguments.
    """

    name = 'submit'
    label = 'Submit'

    @property
    def cache_id(self):
        """:attr:`.TrackerBase.name`"""
        return self.kwargs['tracker'].name

    def initialize(self, *, tracker, tracker_jobs):
        """
        Set internal state

        :param TrackerBase tracker: Return value of :func:`~.trackers.tracker`
        :param TrackerJobsBase tracker_jobs: Instance of
            :attr:`~.base.TrackerBase.TrackerJobs`
        """
        assert isinstance(tracker, trackers.TrackerBase), f'Not a TrackerBase: {tracker!r}'
        assert isinstance(tracker_jobs, trackers.TrackerJobsBase), f'Not a TrackerJobsBase: {tracker_jobs!r}'
        for t in (tracker, tracker_jobs):
            t.signal.register('warning', self.warn)
            t.signal.register('error', self.error)
            t.signal.register('exception', self.exception)
        self._tracker = tracker
        self._tracker_jobs = tracker_jobs
        self._submit_lock = asyncio.Lock()
        for signal, message in (('logging_in', 'Logging in'),
                                ('logged_in', 'Logged in'),
                                ('uploading', 'Uploading'),
                                ('uploaded', 'Uploaded'),
                                ('logging_out', 'Logging out'),
                                ('logged_out', '')):
            self.signal.add(signal)
            self.signal.register(signal, lambda msg=message: setattr(self, 'info', msg))
        self.signal.register('finished', lambda _: setattr(self, 'info', ''))
        self.signal.register('error', lambda _: setattr(self, 'info', ''))

        # Create jobs_after_upload now so they can connect to other jobs,
        # e.g. add_torrent_job needs to register for create_torrent_job's output
        # before create_torrent_job finishes.
        self.jobs_after_upload

        # Start jobs_after_upload only if the upload succeeds. This also works
        # nicely when this job is getting it's output from cache (execute() is
        # not called).
        self.signal.register('output', self._start_jobs_after_upload)

    def _start_jobs_after_upload(self, _):
        for job in self.jobs_after_upload:
            job.start()

    def execute(self):
        """
        Start the asynchronous background task that waits for jobs and wait for it
        to finish
        """
        if not self.is_finished:
            self.attach_task(self._run_jobs())

    async def _run_jobs(self):
        # When all jobs are finished, check again if all jobs are finished. This
        # is necessary because each job can can enable or disable other jobs and
        # self.jobs_before_upload can return a different set of jobs every time.
        # We assume that nothing changes once all jobs are finished.
        jobs_before_upload = self.jobs_before_upload
        while not all(job.is_finished for job in jobs_before_upload):
            _log.debug('Waiting for jobs before upload: %r', [j.name for j in jobs_before_upload])
            await asyncio.gather(*(job.wait() for job in jobs_before_upload))
            jobs_before_upload = self.jobs_before_upload
        _log.debug('Done waiting for jobs before upload: %r', [j.name for j in jobs_before_upload])

        # Wait for any dangling background tasks
        await self._tracker.await_tasks()

        # Don't submit if self._tracker_jobs thinks that's a bad idea.
        if self._tracker_jobs.submission_ok:
            await self._submit()
        self.finish()

    async def _submit(self):
        _log.debug('%s: Submitting', self._tracker.name)
        try:
            self.signal.emit('logging_in')
            await self._tracker.login()
            self.signal.emit('logged_in')
            try:
                self.signal.emit('uploading')
                torrent_page_url = await self._tracker.upload(self._tracker_jobs)
                if torrent_page_url:
                    self.send(torrent_page_url)
                    self.signal.emit('uploaded')
            finally:
                self.signal.emit('logging_out')
                await self._tracker.logout()
                self.signal.emit('logged_out')
        except errors.RequestError as e:
            self.error(e)

    @property
    def jobs_before_upload(self):
        """
        Sequence of jobs to do before submission

        This is the same as :attr:`.TrackerJobsBase.jobs_before_upload` but
        with all `None` values and disabled jobs filtered out.
        """
        return tuple(job for job in self._tracker_jobs.jobs_before_upload
                     if job and job.is_enabled)

    @property
    def jobs_after_upload(self):
        """
        Sequence of jobs to do after successful submission

        This is the same as :attr:`.TrackerJobsBase.jobs_after_upload` but with
        all `None` values and disabled jobs filtered out.
        """
        return tuple(job for job in self._tracker_jobs.jobs_after_upload
                     if job and job.is_enabled)

    @property
    def hidden(self):
        """
        Hide this job if :attr:`~.base.TrackerJobsBase.submission_ok` is falsy

        This allows jobs to prevent submission.

        It also should mean this job is hidden until the submission happens
        because :attr:`~.base.TrackerJobsBase.submission_ok` should be falsy
        until all :attr:`~.base.TrackerJobsBase.jobs_before_upload` finished
        successfully.
        """
        return not self._tracker_jobs.submission_ok

    @property
    def final_job_before_upload(self):
        """
        If submission is prevented by :attr:`~.base.TrackerJobsBase.submission_ok`
        return the last item of
        :attr:`~.base.TrackerJobsBase.jobs_before_upload` if all jobs are
        finished

        Return `None` if :attr:`~.base.TrackerJobsBase.submission_ok` is truthy
        or the final job cannot be determined at the moment.
        """
        if not self._tracker_jobs.submission_ok:
            jobs = self.jobs_before_upload
            # Because jobs can enable/disable each other, we can't know the
            # final job until all jobs are finished.
            if jobs and all(job.is_finished for job in jobs):
                return jobs[-1]

    @property
    def output(self):
        """Output of this job or :attr:`final_job_before_upload` if it is not `None`"""
        final_job_before_upload = self.final_job_before_upload
        if final_job_before_upload:
            return final_job_before_upload.output
        else:
            return super().output

    @property
    def exit_code(self):
        """Exit code of this job or :attr:`final_job_before_upload` if it is not `None`"""
        final_job_before_upload = self.final_job_before_upload
        if final_job_before_upload:
            return final_job_before_upload.exit_code
        else:
            return super().exit_code



"""
Concrete :class:`~.base.TrackerJobsBase` subclass for PTP
"""

from datetime import datetime

from ... import errors, jobs, utils
from ...utils import cached_property
from ...utils.release import ReleaseType
from ..base import TrackerJobsBase
from ._ptp_tags import PTP_TAGS

import logging  # isort:skip
_log = logging.getLogger(__name__)


class PtpTrackerJobs(TrackerJobsBase):

    @cached_property
    def jobs_before_upload(self):
        return (
            # Common background jobs
            self.create_torrent_job,
            self.mediainfo_job,
            self.screenshots_job,
            self.upload_screenshots_job,
            self.ptp_group_id_job,

            # Common interactive jobs
            self.type_job,
            self.imdb_job,
            self.title_job,

            # Jobs that only run if movie exists on PTP and we add a release

            # Jobs that only run if movie does not exists on PTP yet
            self.year_job,
            self.plot_job,
            self.tags_job,
            self.cover_art_job,
        )

    @cached_property
    def type_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('type'),
            label='Type',
            condition=self.make_job_condition('type_job'),
            autodetect=self.autodetect_type,
            choices=(
                ('Feature Film', 'Feature Film'),
                ('Short Film', 'Short Film'),
                ('Miniseries', 'Miniseries'),
                ('Stand-up Comedy', 'Stand-up Comedy'),
                ('Live Performance', 'Live Performance'),
                ('Movie Collection', 'Movie Collection'),
            ),
            **self.common_job_args(),
        )

    def autodetect_type(self, _):
        if self.release_name.type == ReleaseType.season:
            return 'Miniseries'

        first_video = utils.video.first_video(self.content_path)

        # Short film if runtime 45 min or less (Rule 1.1.1)
        if utils.video.duration(first_video) <= 45 * 60:
            return 'Short Film'

    @cached_property
    def imdb_job(self):
        imdb_job = super().imdb_job
        imdb_job.signal.register('output', self._start_getting_ptp_group_id)
        return imdb_job

    @property
    def imdb_id(self):
        if self.imdb_job.is_finished:
            return self.imdb_job.output[0]

    def _start_getting_ptp_group_id(self, _):
        _log.debug('Starting %r', self.ptp_group_id_job)
        self.ptp_group_id_job.start()

    @cached_property
    def ptp_group_id_job(self):
        return jobs.custom.CustomJob(
            name=self.get_job_name('ptp-group-id'),
            label='PTP Group ID',
            condition=self.make_job_condition('ptp_group_id_job'),
            prejobs=(self.imdb_job,),
            worker=self.fetch_ptp_group_id,
            catch=(errors.RequestError,),
            **self.common_job_args(),
        )

    async def fetch_ptp_group_id(self, _):
        assert self.imdb_job.is_finished
        group_id = await self.tracker.get_ptp_group_id_by_imdb_id(self.imdb_id)
        _log.debug('Group ID: %r', group_id)
        return '' if group_id is None else group_id

    @property
    def ptp_group_id(self):
        """
        PTP group ID if :attr:`ptp_group_id_job` is finished and group ID was found,
        `None` otherwise
        """
        if self.ptp_group_id_job.is_finished:
            if self.ptp_group_id_job.output:
                return self.ptp_group_id_job.output[0]

    @cached_property
    def title_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('title'),
            label='Title',
            condition=self.make_job_condition('title_job'),
            prejobs=(self.imdb_job,),
            text=self.fetch_title,
            normalizer=self.normalize_title,
            validator=self.validate_title,
            **self.common_job_args(),
        )

    async def fetch_title(self):
        assert self.imdb_job.is_finished
        _log.debug('!!!!!! Filling in new title')
        info = await self.tracker.get_ptp_metadata(self.imdb_id)
        _log.debug('PTP metadata: %r', info)
        return info['title']

    def normalize_title(self, text):
        return text.strip()

    def validate_title(self, text):
        if not text:
            raise ValueError('Title must not be empty')

    @cached_property
    def year_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('year'),
            label='Year',
            condition=self.make_job_condition('year_job'),
            prejobs=(self.imdb_job,),
            text=self.fetch_year,
            normalizer=self.normalize_year,
            validator=self.validate_year,
            **self.common_job_args(),
        )

    async def fetch_year(self):
        assert self.imdb_job.is_finished
        _log.debug('!!!!!! Filling in new year')
        info = await self.tracker.get_ptp_metadata(self.imdb_id)
        _log.debug('PTP metadata: %r', info)
        return info['year']

    def normalize_year(self, text):
        return text.strip()

    def validate_year(self, text):
        try:
            year = int(text)
        except ValueError:
            raise ValueError('Year is not a number')
        else:
            if not 1800 < year < datetime.now().year + 10:
                raise ValueError('Year is not reasonable')

    @cached_property
    def tags_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('tags'),
            label='Tags',
            condition=self.make_job_condition('tags_job'),
            prejobs=(self.imdb_job,),
            text=self.fetch_tags,
            normalizer=self.normalize_tags,
            validator=self.validate_tags,
            **self.common_job_args(),
        )

    async def fetch_tags(self):
        assert self.imdb_job.is_finished
        _log.debug('!!!!!! Filling in new tags')
        info = await self.tracker.get_ptp_metadata(self.imdb_id)
        _log.debug('PTP metadata: %r', info)
        return info['tags']

    def normalize_tags(self, text):
        tags = [tag.strip() for tag in text.split(',')]
        deduped = list(dict.fromkeys(tags))
        return ', '.join(tag for tag in deduped if tag)

    def validate_tags(self, text):
        for tag in text.split(','):
            if tag.strip() not in PTP_TAGS:
                raise ValueError(f'Tag is not valid: {tag}')

    @cached_property
    def cover_art_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('cover_art'),
            label='Cover Art',
            condition=self.make_job_condition('cover_art_job'),
            prejobs=(self.imdb_job,),
            text=self.fetch_cover_art,
            autofinish=True,
            **self.common_job_args(),
        )

    async def fetch_cover_art(self):
        assert self.imdb_job.is_finished
        _log.debug('!!!!!! Filling in new cover_art')
        info = await self.tracker.get_ptp_metadata(self.imdb_id)
        _log.debug('PTP metadata: %r', info)
        return info['art']

    @cached_property
    def plot_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('plot'),
            label='Plot',
            condition=self.make_job_condition('plot_job'),
            prejobs=(self.imdb_job,),
            text=self.fetch_plot,
            normalizer=self.normalize_plot,
            validator=self.validate_plot,
            autofinish=True,
            **self.common_job_args(),
        )

    async def fetch_plot(self):
        assert self.imdb_job.is_finished
        return await self.tracker.get_ptp_metadata(self.imdb_id, key='plot')

    def normalize_plot(self, text):
        return text.strip()

    def validate_plot(self, text):
        if not text:
            raise ValueError('Plot must not be empty')

    @property
    def post_data(self):
        post_data = self._post_data_common

        _log.debug('PTP group ID: %r', self.ptp_group_id)
        if self.ptp_group_id:
            _log.debug('Adding format to existing group')
            post_data.update(self._post_data_add_format)
        else:
            _log.debug('Creating new group')
            post_data.update(self._post_data_add_movie)

        return post_data

    @property
    def _post_data_common(self):
        return {
            'type': self.get_job_attribute(self.type_job, 'choice'),
            # 'remaster_year': releaseInfo.RemasterYear,
            # 'remaster_title': releaseInfo.RemasterTitle,
            # 'codec': 'Other',  # Sending the codec as custom.
            # 'other_codec': releaseInfo.Codec,
            # 'container': 'Other',  # Sending the container as custom.
            # 'other_container': releaseInfo.Container,
            # 'resolution': releaseInfo.ResolutionType,
            # 'other_resolution': releaseInfo.Resolution,
            # 'source': 'Other',  # Sending the source as custom.
            # 'other_source': releaseInfo.Source,
            # 'release_desc': releaseDescription,
            # 'nfo_text': releaseInfo.Nfo,
            # 'subtitles[]': releaseInfo.Subtitles,
            # 'trumpable[]': releaseInfo.Trumpable,
        }

    # if releaseInfo.Source == "Other":
    #     paramList["other_source"]: releaseInfo.SourceOther

    # # personal rip only needed if it is specified
    # if releaseInfo.PersonalRip:
    #     paramList.update({"internalrip": "on"})

    # # scene only needed if it is specified
    # if releaseInfo.SceneRelease:
    #     paramList.update({"scene": "on"})
    # # other category is only needed if it is specified
    # if releaseInfo.SpecialRelease:
    #     paramList.update({"special": "on"})
    # # remaster is only needed if it is specified
    # if len(releaseInfo.RemasterYear) > 0 or len(releaseInfo.RemasterTitle) > 0:
    #     paramList.update({"remaster": "on"})
    #     }

    @property
    def _post_data_add_format(self):
        return {
            'groupid': self.ptp_group_id,
        }

    @property
    def _post_data_add_movie(self):
        post_data = {
            'imdb': self.tracker.normalize_imdb_id(self.imdb_id),
            'title': self.get_job_output(self.title_job, slice=0),
            'year': self.get_job_output(self.year_job, slice=0),
            'album_desc': self.get_job_output(self.plot_job, slice=0),
            'tags': self.get_job_output(self.tags_job, slice=0),
            # 'trailer': releaseInfo.YouTubeId,
            'image': self.get_job_output(self.cover_art_job, slice=0),
        }

        # TODO: Send artists
        # # ???: For some reason PtpUploader uses "artist" and "importance" in the
        # #      POST data while the website form uses "artists", "importances"
        # #      and "roles".
        # artists = self.get_job_output(self.artists_job):
        # if artists:
        #     post_data['artist'] = []
        #     post_data['importance'] = []
        #     for name in artists:
        #         # `importance` is the artist type:
        #         # 1: Director
        #         # 2: Writer
        #         # 3: Producer
        #         # 4: Composer
        #         # 5: Actor
        #         # 6: Cinematographer
        #         post_data['importance'].append('1')
        #         post_data['artist'].append(name)
        #         post_data['role'].append(name)

        return post_data

    @property
    def torrent_filepath(self):
        return self.get_job_output(self.create_torrent_job, slice=0)



"""
Create torrent file
"""

import collections
import datetime
import errno
import os
import time

from .. import __project_name__, __version__, constants, errors, utils
from . import LazyModule, fs, types

torf = LazyModule(module='torf', namespace=globals())


SKIP_SEARCHING = 'skip_searching'
"""
Return value for the `progress_callback` to stop searching for a reusable
torrent (see :func:`create`)
"""


def create(*, content_path, announce, source, torrent_path,
           use_cache=True, exclude=(), reuse_torrent_path=None,
           piece_size_min=None, piece_size_max=None,
           init_callback, progress_callback):
    """
    Generate and write torrent file

    :param str content_path: Path to the torrent's payload
    :param str announce: Announce URL
    :param str source: Value of the ``source`` field in the torrent. This makes
        the torrent unique for each tracker to avoid cross-seeding issues, so it
        is usually the tracker's abbreviated name.
    :param str torrent_path: Path of the generated torrent file
    :param int piece_size_min: Minimum piece size
    :param int piece_size_max: Maximum piece size
    :param init_callback: Callable that is called once before torrent generation
        commences. It gets `content_path` as a tree where each node is a tuple
        in which the first item is the directory name and the second item is a
        sequence of `(file_name, file_size)` tuples.

        Example:

        .. code::

           ('Parent',
               ('Foo', (
                   ('Picture.jpg', 82489),
                   ('Music.mp3', 5315672),
                   ('More files', (
                       ('This.txt', 57734),
                       ('And that.txt', 184),
                       ('Also some of this.txt', 88433),
                   )),
               )),
               ('Bar', (
                   ('Yee.mp4', 288489392),
                   ('Yah.mkv', 3883247384),
               )),
           )
    :param progress_callback: Callable that is called at regular intervals with
        a :class:`CreateTorrentProgress` or :class:`FindTorrentProgress` object
        as a positional argument
    :param exclude: Sequence of regular expressions that are matched against
        file system paths. Matching files are not included in the torrent.
    :param bool use_cache: Whether to get piece hashes from previously created
        torrents or from `reuse_torrent_path`
    :param reuse_torrent_path: Path to existing torrent file to get hashed
        pieces and piece size from. If the given torrent file doesn't match the
        files in the torrent we want to create, hash the pieces normally.

        If this is a directory, search it recursively for ``*.torrent`` files
        and use the first one that matches.

        Non-existing or otherwise unreadable paths as well as falsy values
        (e.g. ``""`` or `None`) are silently ignored.

        If this is a sequence, its items are expected to be directory or file
        paths and handled as described above.

    Callbacks can cancel the torrent creation by returning `True` or any other
    truthy value. If `progress_callback` returns :data:`SKIP_SEARCHING`, the
    search for a reusable torrent is cancelled and pieces are hashed normally.

    :raise TorrentError: if anything goes wrong

    :return: `torrent_path` or `None` if cancelled
    """
    if not announce:
        raise errors.TorrentError('Announce URL is empty')
    if not source:
        raise errors.TorrentError('Source is empty')

    # Create Torrent object
    torrent = _get_torrent(
        content_path=content_path,
        exclude=exclude,
        announce=announce,
        source=source,
        piece_size_min=piece_size_min,
        piece_size_max=piece_size_max,
    )
    # Report files with `exclude` applied
    cancelled = init_callback(_make_file_tree(torrent.filetree))
    if cancelled:
        return None

    if use_cache:
        # Try to get piece hashes from existing torrent
        cancelled = _find_hashes(
            torrent=torrent,
            reuse_torrent_path=reuse_torrent_path,
            callback=progress_callback,
        )
        if cancelled and cancelled != SKIP_SEARCHING:
            return None

    if not torrent.is_ready:
        # Hash pieces
        cancelled = _generate_hashes(
            torrent=torrent,
            callback=progress_callback,
        )
        if cancelled:
            return None

    # Write generic torrent so we can reuse the hashes in the future
    _store_generic_torrent(torrent)

    # Write torrent to `torrent_path`
    _write_torrent_path(torrent, torrent_path)

    return torrent_path


def _get_torrent(*, content_path, exclude, announce, source, piece_size_min, piece_size_max):
    try:
        return torf.Torrent(
            path=content_path,
            exclude_regexs=exclude,
            piece_size_min=piece_size_min,
            piece_size_max=piece_size_max,
            trackers=((announce,),),
            source=source,
            private=True,
            created_by=f'{__project_name__} {__version__}',
            creation_date=time.time(),
        )
    except torf.TorfError as e:
        raise errors.TorrentError(str(e))


def _generate_hashes(*, torrent, callback):
    wrapped_callback = _CreateTorrentCallback(callback)
    try:
        torrent.generate(
            callback=wrapped_callback,
            interval=1.0,
        )
    except torf.TorfError as e:
        raise errors.TorrentError(str(e))
    else:
        return wrapped_callback.return_value


def _find_hashes(*, torrent, reuse_torrent_path, callback):
    wrapped_callback = _FindTorrentCallback(callback)
    try:
        torrent.reuse(
            _get_reuse_torrent_paths(torrent, reuse_torrent_path),
            callback=wrapped_callback,
            interval=1.0,
        )
    except torf.TorfError as e:
        raise errors.TorrentError(str(e))
    else:
        return wrapped_callback.return_value


def _get_reuse_torrent_paths(torrent, reuse_torrent_path):
    reuse_torrent_paths = []
    if reuse_torrent_path:
        if isinstance(reuse_torrent_path, str):
            reuse_torrent_paths.append(reuse_torrent_path)
        elif isinstance(reuse_torrent_path, collections.abc.Iterable):
            reuse_torrent_paths.extend(p for p in reuse_torrent_path if p)
        else:
            raise ValueError(f'Invalid reuse_torrent_path: {reuse_torrent_path!r}')

    generic_torrent_path = _get_generic_torrent_path(torrent=torrent, create_directory=False)
    reuse_torrent_paths.insert(0, generic_torrent_path)
    return reuse_torrent_paths


def _store_generic_torrent(torrent):
    generic_torrent = torf.Torrent(
        private=True,
        created_by=f'{__project_name__} {__version__}',
        creation_date=time.time(),
        comment='This torrent is used to cache previously hashed pieces.',
    )
    _copy_torrent_info(torrent, generic_torrent)
    generic_torrent_path = _get_generic_torrent_path(generic_torrent, create_directory=True)
    generic_torrent.write(generic_torrent_path, overwrite=True)


def _write_torrent_path(torrent, torrent_path):
    try:
        torrent.write(torrent_path, overwrite=True)
    except torf.TorfError as e:
        raise errors.TorrentError(str(e))


def _get_generic_torrent_path(torrent, create_directory=True):
    directory = constants.GENERIC_TORRENTS_DIRPATH
    if create_directory:
        try:
            fs.mkdir(directory)
        except errors.ContentError as e:
            raise errors.TorrentError(f'{directory}: {e}')

    cache_id = _get_torrent_id(torrent)
    filename = f'{torrent.name}.{cache_id}.torrent'
    return os.path.join(directory, filename)


def _get_torrent_id_info(torrent):
    return {
        'name': torrent.name,
        'files': tuple((str(f), f.size) for f in torrent.files),
    }


def _get_torrent_id(torrent):
    return utils.semantic_hash(_get_torrent_id_info(torrent))


def _copy_torrent_info(from_torrent, to_torrent):
    from_info, to_info = from_torrent.metainfo['info'], to_torrent.metainfo['info']
    to_info['pieces'] = from_info['pieces']
    to_info['piece length'] = from_info['piece length']
    to_info['name'] = from_info['name']
    if 'length' in from_info:
        to_info['length'] = from_info['length']
    else:
        to_info['files'] = from_info['files']


def _make_file_tree(filetree):
    files = []
    for name,file in filetree.items():
        if isinstance(file, collections.abc.Mapping):
            subtree = _make_file_tree(file)
            files.append((name, subtree))
        else:
            files.append((name, file.size))
    return tuple(files)


class _CallbackBase:
    def __init__(self, callback):
        self._callback = callback
        self._progress_samples = []
        self._time_started = time.time()
        self.return_value = None

    def __call__(self, progress):
        self.return_value = return_value = self._callback(progress)
        return return_value

    def _calculate_info(self, items_done, items_total):
        time_now = time.time()
        percent_done = items_done / items_total * 100
        seconds_elapsed = time_now - self._time_started
        items_per_second = 0
        items_remaining = items_total - items_done
        seconds_remaining = 0

        self._add_sample(time_now, items_done)

        # Estimate how long torrent creation will take
        samples = self._progress_samples
        if len(samples) >= 2:
            # Calculate the difference between each pair of samples
            diffs = [
                b[1] - a[1]
                for a, b in zip(samples[:-1], samples[1:])
            ]
            items_per_second = self._get_average(diffs, weight_factor=1.1)
            if items_per_second > 0:
                seconds_remaining = items_remaining / items_per_second

        seconds_total = seconds_elapsed + seconds_remaining
        time_finished = self._time_started + seconds_total

        return {
            'percent_done': percent_done,
            'items_remaining': items_remaining,
            'items_per_second': items_per_second,
            'seconds_elapsed': datetime.timedelta(seconds=seconds_elapsed),
            'seconds_remaining': datetime.timedelta(seconds=seconds_remaining),
            'seconds_total': datetime.timedelta(seconds=seconds_total),
            'time_finished': datetime.datetime.fromtimestamp(time_finished),
            'time_started': datetime.datetime.fromtimestamp(self._time_started),
        }

    def _add_sample(self, time_now, items_done):
        def get_sample_age(sample):
            time_sample = sample[0]
            return time_now - time_sample

        samples = self._progress_samples
        samples.append((time_now, items_done))

        # Prune samples older than 10 seconds
        while samples and get_sample_age(samples[0]) > 10:
            del samples[0]

    def _get_average(self, samples, weight_factor, get_value=lambda sample: sample):
        # Give recent samples more weight than older samples
        # https://en.wikipedia.org/wiki/Moving_average
        weights = []
        for _ in range(len(samples)):
            try:
                weight = weights[-1]
            except IndexError:
                weight = 1
            weights.append(weight * weight_factor)

        return sum(
            get_value(sample) * weight
            for sample, weight in zip(samples, weights)
        ) / sum(weights)


class _CreateTorrentCallback(_CallbackBase):
    def __call__(self, torrent, filepath, pieces_done, pieces_total):
        info = self._calculate_info(pieces_done, pieces_total)
        piece_size = torrent.piece_size
        bytes_per_second = types.Bytes(info['items_per_second'] * piece_size)
        progress = CreateTorrentProgress(
            pieces_done=pieces_done,
            pieces_total=pieces_total,
            percent_done=info['percent_done'],
            bytes_per_second=bytes_per_second,
            piece_size=piece_size,
            total_size=torrent.size,
            filepath=filepath,
            seconds_elapsed=info['seconds_elapsed'],
            seconds_remaining=info['seconds_remaining'],
            seconds_total=info['seconds_total'],
            time_finished=info['time_finished'],
            time_started=info['time_started'],
        )
        return super().__call__(progress)


class _FindTorrentCallback(_CallbackBase):
    def __call__(self, torrent, filepath, files_done, files_total, status, exception):
        info = self._calculate_info(files_done, files_total)

        # Ignore "No such file or directory". This should only happen if the
        # generic torrent does not exist yet. For all other paths, torrent files
        # are collected from traversing directories.
        if isinstance(exception, torf.ReadError) and exception.errno == errno.ENOENT:
            exception = None

        if status is True:
            status = 'hit'
        elif status is False:
            status = 'miss'
        elif status is None:
            status = 'verifying'

        progress = FindTorrentProgress(
            files_done=files_done,
            files_total=files_total,
            percent_done=info['percent_done'],
            files_per_second=info['items_per_second'],
            filepath=filepath,
            status=status,
            exception=errors.TorrentError(str(exception)) if exception else None,
            seconds_elapsed=info['seconds_elapsed'],
            seconds_remaining=info['seconds_remaining'],
            seconds_total=info['seconds_total'],
            time_finished=info['time_finished'],
            time_started=info['time_started'],
        )
        return super().__call__(progress)


class CreateTorrentProgress(collections.namedtuple(
    typename='CreateTorrentProgress',
    field_names=(
        'bytes_per_second',
        'filepath',
        'percent_done',
        'piece_size',
        'pieces_done',
        'pieces_total',
        'seconds_elapsed',
        'seconds_remaining',
        'seconds_total',
        'time_finished',
        'time_started',
        'total_size',
    ),
)):
    """
    :func:`~.collections.namedtuple` with these attributes:

        - ``bytes_per_second`` (:class:`~.types.Bytes`)
        - ``filepath`` (:class:`str`)
        - ``percent_done`` (:class:`float`)
        - ``piece_size`` (:class:`~.types.Bytes`)
        - ``pieces_done`` (:class:`int`)
        - ``pieces_total`` (:class:`int`)
        - ``seconds_elapsed`` (:class:`~.datetime.datetime.timedelta`)
        - ``seconds_remaining`` (:class:`~.datetime.datetime.timedelta`)
        - ``seconds_total`` (:class:`~.datetime.datetime.timedelta`)
        - ``time_finished`` (:class:`~.datetime.datetime.datetime`)
        - ``time_started`` (:class:`~.datetime.datetime.datetime`)
        - ``total_size`` (:class:`~.types.Bytes`)
    """


class FindTorrentProgress(collections.namedtuple(
    typename='CreateTorrentProgress',
    field_names=(
        'exception',
        'filepath',
        'files_done',
        'files_per_second',
        'files_total',
        'percent_done',
        'seconds_elapsed',
        'seconds_remaining',
        'seconds_total',
        'status',
        'time_finished',
        'time_started',
    ),
)):
    """
    :func:`~.collections.namedtuple` with these attributes:

        - ``exception`` (:class:`~.errors.TorrentError` or `None`)
        - ``filepath`` (:class:`str`)
        - ``files_done`` (:class:`int`)
        - ``files_per_second`` (:class:`int`)
        - ``files_total`` (:class:`int`)
        - ``percent_done`` (:class:`float`)
        - ``seconds_elapsed`` (:class:`~.datetime.datetime.timedelta`)
        - ``seconds_remaining`` (:class:`~.datetime.datetime.timedelta`)
        - ``seconds_total`` (:class:`~.datetime.datetime.timedelta`)
        - ``status`` (``hit``, ``miss`` or ``verifying``)
        - ``time_finished`` (:class:`~.datetime.datetime.datetime`)
        - ``time_started`` (:class:`~.datetime.datetime.datetime`)
    """



"""
String formatting and parsing
"""

import importlib
import inspect
import re
import sys

import logging  # isort:skip
_log = logging.getLogger(__name__)


def group(name):
    """
    Return character group, e.g. "digits"

    This function simply returns an attribute from the built-in :mod:`string`
    module.

    :raise ValueError: if the built-in :mod:`string` module doesn't have
        attribute `name`
    """
    string_module = importlib.import_module('string')
    try:
        return getattr(string_module, name)
    except AttributeError:
        raise ValueError(f'No such character group: {name!r}')


def autodecode(bytes):
    """
    Decode `bytes` with autodetected encoding

    :raise ValueError: if autodetection fails

    :return: Decoded string
    :rtype: str
    """
    import chardet
    info = chardet.detect(bytes)
    encoding = info.get('encoding', None)
    if encoding:
        _log.debug('Decoding %r with %r', bytes[:100], encoding)
        return str(bytes, encoding)
    else:
        raise ValueError(f'Unable to autodetect encoding: {bytes!r}')


_capitalize_regex = re.compile(r'(\s*)(\S+)(\s*)')

def capitalize(text):
    """
    Capitalize each word in `text`

    Unlike :meth:`str.title`, only words at in front of a space or at the
    beginning of `text` are capitalized.
    """
    return ''.join(
        match.group(1) + match.group(2).capitalize() + match.group(3)
        for match in re.finditer(_capitalize_regex, text)
    )


def star_rating(rating, max_rating=10):
    """
    Return star rating string with the characters "★" (U+2605), "⯪" (U+2BEA) and
    "☆" (U+2605)

    :param float,int rating: Number between 0 and `max_rating`
    :param float,int max_rating: Maximum rating
    """
    import math
    rating = min(max_rating, max(0, rating))
    left = '\u2605' * math.floor(rating)
    if rating >= max_rating:
        middle = ''
    # Avoid floating point precision issues by rounding to 1 digit after comma
    elif round(rating % 1, 1) <= 0.3:
        middle = '\u2606'  # Empty star
    elif round(rating % 1, 1) < 0.7:
        middle = '\u2bea'  # Half star
    else:
        middle = '\u2605'  # Full star
    right = '\u2606' * (math.ceil(max_rating - rating) - 1)
    return f'{left}{middle}{right}'


if sys.version_info >= (3, 9, 0):
    def remove_prefix(string, prefix):
        return string.removeprefix(prefix)

    def remove_suffix(string, suffix):
        return string.removesuffix(suffix)

else:
    def remove_prefix(string, prefix):
        if string.startswith(prefix):
            return string[len(prefix):]
        else:
            return string

    def remove_suffix(string, suffix):
        if string.endswith(suffix):
            return string[:-len(suffix)]
        else:
            return string


def evaluate_fstring(template):
    """
    Deferred f-string evaluation

    Example:

    >>> template = 'Evaluate this: {sum((1, 2, 3)) * 2}'
    >>> evaluate_fstring(template)
    12

    References:

        https://pypi.org/project/f-yeah/
        https://stackoverflow.com/a/42497694
    """
    parent_frame = inspect.stack()[1].frame
    fstring = 'f' + repr(template)
    return eval(fstring, parent_frame.f_globals, parent_frame.f_locals)



"""
Concrete :class:`~.base.TrackerConfigBase`, :class:`~.base.TrackerJobsBase`
and :class:`~.base.TrackerBase` subclasses for NBL
"""

from .config import NblTrackerConfig
from .jobs import NblTrackerJobs
from .tracker import NblTracker



from base64 import b64decode

from .. import html, http
from . import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


class CorruptnetApi(base.PredbApiBase):
    name = 'corruptnet'
    label = 'pre.corrupt-net.org'

    default_config = {}

    _url_base = 'https://' + b64decode('cHJlLmNvcnJ1cHQtbmV0Lm9yZw==').decode('ascii')
    _search_url = f'{_url_base}/search.php'

    def _join_keywords(self, keywords, group):
        kws = ' '.join((kw.strip() for kw in keywords))
        if group:
            kws += f' group:{group.strip()}'
        return kws

    # This seems to be specified by the server?
    _results_per_page = 50

    # Seems reasonable?
    _max_pages = 10

    async def _search(self, keywords, group):
        combined_results = []
        max_page_value = self._max_pages * self._results_per_page
        for page in range(0, max_page_value, self._results_per_page):
            results = list(await self._request_page(keywords, group, page))
            combined_results.extend(results)
            if len(results) < self._results_per_page:
                break
        return combined_results

    async def _request_page(self, keywords, group, page):
        params = {
            'search': self._join_keywords(keywords, group),
            # This should be the current time in milliseconds, but if we do
            # this, nothing is cached, so we set this to 0.
            'ts': 0,
            # Not sure what these are
            'pretimezone': 0,
            'timezone': 0,
        }
        if page > 0:
            params['page'] = page

        _log.debug('%s search: %r, %r', self.label, self._search_url, params)
        response = await http.get(self._search_url, params=params, cache=True, verify=False)
        return self._yield_release_names(response)

    def _yield_release_names(self, response):
        doc = html.parse(response)
        for tr in doc.find_all('tr'):
            # Release name is in second column
            td_release_name = tr.select_one('td:nth-of-type(2)')
            if td_release_name:
                release_name = td_release_name.get_text().strip()
                yield release_name

    async def _release_files(self, release_name):
        raise NotImplementedError()



"""
Scene release search and verification
"""

from ... import utils
from .. import release
from ..types import SceneCheckResult
from . import corruptnet, predbde, predbovh, srrdb
from .base import PredbApiBase
from .common import (assert_not_abbreviated_filename, is_abbreviated_filename,
                     is_mixed_season_pack)
from .multi import MultiPredbApi
from .query import SceneQuery


def predbs():
    """Return list of :class:`~.PredbApiBase` subclasses"""
    return utils.subclasses(PredbApiBase, utils.submodules(__package__))


def predb(name, config=None):
    """
    Create :class:`~.PredbApiBase` instance

    :param str name: Name of the scene release database. A subclass of
        :class:`~.PredbApiBase` with the same :attr:`~.PredbApiBase.name`
        must exist in one of this package's submodules.
    :param dict config: User configuration passed to the subclass specified by
        `name`

    :raise ValueError: if no matching subclass can be found

    :return: :class:`~.PredbApiBase` instance
    """
    for predb in predbs():
        if predb.name == name:
            return predb(config=config)
    raise ValueError(f'Unsupported scene release database: {name}')


def predb_names():
    """Return sequence of valid `name` arguments for :func:`.predb`"""
    return sorted(utils.CaseInsensitiveString(cls.name) for cls in predbs())



"""
Concrete :class:`~.base.TrackerBase` subclass for bB
"""

import asyncio
import re
import urllib

from ... import __project_name__, errors
from ...utils import html, http
from ..base import TrackerBase
from .config import BbTrackerConfig
from .jobs import BbTrackerJobs

import logging  # isort:skip
_log = logging.getLogger(__name__)


class BbTracker(TrackerBase):
    name = 'bb'
    label = 'bB'

    TrackerConfig = BbTrackerConfig
    TrackerJobs = BbTrackerJobs

    _url_path = {
        'login': '/login.php',
        'logout': '/logout.php',
        'upload': '/upload.php',
        'torrent': '/torrents.php',
    }

    _max_login_attempts = 30
    _login_retry_delay = 1
    _login_retry_delay_max = 3

    async def login(self):
        if not self.is_logged_in:
            if not self.options.get('username'):
                raise errors.RequestError('Login failed: No username configured')
            elif not self.options.get('password'):
                raise errors.RequestError('Login failed: No password configured')

            login_url = urllib.parse.urljoin(
                self.options['base_url'],
                self._url_path['login'],
            )

            doc = html.parse('')
            while await self._work_around_login_bug(doc):
                _log.debug('%s: Logging in as %r', self.name, self.options['username'])
                response = await http.post(
                    url=login_url,
                    user_agent=True,
                    data={
                        'username': self.options['username'],
                        'password': self.options['password'],
                        'login': 'Log In!',
                    },
                )
                doc = html.parse(response)
                self._raise_login_error(doc)

            self._store_auth_token(doc)

    def _raise_login_error(self, doc):
        error_tag = doc.find('font', color='red')
        if error_tag:
            raise errors.RequestError(f'Login failed: {error_tag.string.strip()}')
        elif doc.find('form', id='loginform'):
            html.dump(doc, 'login.html')
            raise errors.RequestError('Login failed: No error message found. See login.html.')

    async def _work_around_login_bug(self, doc):
        # Login sometimes fails and redirects to main page without logging in
        if not self._get_auth_token(doc):
            if not hasattr(self, '_login_attempts'):
                self._login_attempts = -1
            self._login_attempts += 1

            if self._login_attempts < self._max_login_attempts:
                if self._login_attempts:
                    self._login_retry_delay = min(self._login_retry_delay * 1.2,
                                                  self._login_retry_delay_max)
                assert type(self)._login_retry_delay == 1
                if self._login_attempts > 1:
                    _log.debug(f'Encountered login bug; retrying in {round(self._login_retry_delay)}s')
                    await asyncio.sleep(round(self._login_retry_delay))
                return True
            else:
                _log.debug(f'Encountered login bug; giving up after {self._login_attempts} attempts')
                raise errors.RequestError('Login failed: Giving up after encountering '
                                          f'login bug {self._login_attempts} times')
        return False

    _auth_token_regex = re.compile(r'logout\.php\?.*\bauth=([0-9a-fA-F]+)')

    @classmethod
    def _get_auth_token(cls, doc):
        logout_link_tag = doc.find('a', href=cls._auth_token_regex)
        if logout_link_tag:
            logout_link_href = logout_link_tag['href']
            if logout_link_href:
                match = cls._auth_token_regex.search(logout_link_href)
                if match:
                    return match.group(1)

    def _store_auth_token(self, doc):
        auth_token = self._get_auth_token(doc)
        if not auth_token:
            raise RuntimeError('Failed to find authentication token')
        else:
            self._auth_token = auth_token

    @property
    def is_logged_in(self):
        return bool(getattr(self, '_auth_token', False))

    async def logout(self):
        if self.is_logged_in:
            _log.debug('%s: Logging out', self.name)
            logout_url = urllib.parse.urljoin(
                self.options['base_url'],
                self._url_path['logout'],
            )
            params = {'auth': self._auth_token}
            delattr(self, '_auth_token')
            await http.get(url=logout_url, params=params, user_agent=True)

    async def get_announce_url(self):
        if self.options.get('announce_url'):
            return self.options['announce_url']
        else:
            await self.login()
            try:
                url = urllib.parse.urljoin(
                    self.options['base_url'],
                    self._url_path['upload'],
                )
                _log.debug('%s: Getting announce URL from %s', self.name, url)
                response = await http.get(url, cache=False, user_agent=True)
                doc = html.parse(response)
                announce_url_tag = doc.find('input', value=re.compile(r'^https?://.*/announce$'))
                if announce_url_tag:
                    return announce_url_tag['value']
                else:
                    cmd = f'{__project_name__} set trackers.{self.name}.announce_url <YOUR URL>'
                    raise errors.RequestError(f'Failed to find announce URL - set it manually: {cmd}')
            finally:
                await self.logout()

    async def upload(self, tracker_jobs):
        _log.debug('Uploading %r', tracker_jobs.post_data)
        upload_url = urllib.parse.urljoin(self.options['base_url'], self._url_path['upload'])
        response = await http.post(
            url=upload_url,
            cache=False,
            user_agent=True,
            follow_redirects=False,
            files={'file_input': {
                'file': tracker_jobs.torrent_filepath,
                'mimetype': 'application/octet-stream',
            }},
            data=tracker_jobs.post_data,
        )

        # Upload response should redirect to torrent page via "Location" header
        torrent_page_url = urllib.parse.urljoin(
            self.options['base_url'],
            response.headers.get('Location', ''),
        )
        if urllib.parse.urlparse(torrent_page_url).path == self._url_path['torrent']:
            return str(torrent_page_url)
        else:
            _log.debug('Unexpected torrent page URL: %r', torrent_page_url)
            doc = html.parse(response)

            # Try to find error message
            error_tag = doc.find('p', style=re.compile(r'color: red'))
            try:
                error_msg = error_tag.string.strip()
                if error_msg:
                    raise errors.RequestError(f'Upload failed: {error_msg}')
            except AttributeError:
                pass

            # Try to find different type of error message
            # (e.g. "The exact same torrent file already exists on the site!")
            try:
                error_header_tag = doc.find('h2', string=re.compile(r'^\s*Error\s*$'))
                parent_tag = error_header_tag.parent
                error_msg_tag = parent_tag.find('p')
                error_msg = error_msg_tag.string.strip()
                if error_msg:
                    raise errors.RequestError(f'Upload failed: {error_msg}')
            except AttributeError:
                pass

            # Try to find warning (e.g. "Your torrent has been uploaded however,
            # you must re-download the torrent file from <a
            # href="torrents.php?id=123456">here</a>, because of the following
            # warnings: The torrent did not have the correct announce URL."
            try:
                warning_header_tag = doc.find('h1', string=re.compile(r'^\s*Warning\s*$'))
                content_tag = warning_header_tag.parent
                warnings = [
                    warning_tag.get_text().strip()
                    for warning_tag in content_tag.find_all('li')
                ]
                if warnings:
                    for warning in warnings:
                        self.warn(warning)
                    return None
            except AttributeError:
                pass

            # Unable to find error message
            html.dump(response, 'upload.html')
            raise RuntimeError('Failed to find error message. See upload.html.')



import re
from base64 import b64decode

from .. import http
from . import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


class SrrdbApi(base.PredbApiBase):
    name = 'srrdb'
    label = 'srrDB'

    default_config = {}

    _url_base = b64decode('YXBpLnNycmRiLmNvbQ==').decode('ascii')
    _search_url = f'https://{_url_base}/v1/search'
    _details_url = f'https://{_url_base}/v1/details'

    _keyword_separators_regex = re.compile(r'[^a-zA-Z0-9]')

    # We can skip over 50_000 results and get 1000 results per request
    # https://www.srrdb.com/help
    _page_size = 1000
    _max_skip = 50_000

    async def _search(self, keywords, group):
        combined_results = []
        for skip in range(0, self._max_skip + 1, self._page_size):
            keywords_path = self._get_keywords_path(keywords, group, skip)
            results = await self._request_page(keywords_path)
            combined_results.extend(results)
            if len(results) < self._page_size:
                # We didn't get a full page, so we assume this is the last page
                break

        return combined_results

    async def _request_page(self, keywords_path):
        search_url = f'{self._search_url}/{keywords_path}'
        _log.debug('Search URL: %s', search_url)
        response = (await http.get(search_url, cache=True)).json()
        results = response.get('results', [])
        return tuple(r['release'] for r in results)

    def _get_keywords_path(self, keywords, group, skip):
        def sanitize_keyword(kw):
            return kw.lower()

        keywords_sanitized = [
            sanitize_keyword(kw)
            for keyword in keywords
            for kw in self._keyword_separators_regex.split(keyword)
        ]
        if group:
            keywords_sanitized.append(f'group:{sanitize_keyword(group)}')

        # Get most recent results
        keywords_sanitized.append('order:date-desc')

        # Skip over `skip` results
        assert (isinstance(skip, int) and 0 <= skip <= self._max_skip), skip
        keywords_sanitized.append(f'skipr:{skip}.{self._page_size}')

        return '/'.join(keywords_sanitized)

    async def _release_files(self, release_name):
        """
        Map file names to dictionaries with the keys ``release_name``,
        ``file_name``, ``size`` and ``crc``

        If no files for `release_name` are found, return an empty :class:`dict`.

        :param str release_name: Exact name of the release

        :raise RequestError: if request fails
        """
        details_url = f'{self._details_url}/{release_name}'
        _log.debug('Details URL: %s', details_url)
        response = await http.get(details_url, cache=True)
        # Response may be empty string
        if response:
            # Response may be empty list
            info = response.json()
            if info:
                # If info is not an empty list, it should be a dictionary
                files = info['archived-files']
                release_name = info['name']
                return {
                    f['name']: {
                        'release_name': release_name,
                        'file_name': f['name'],
                        'size': f['size'],
                        'crc': f['crc'],
                    }
                    for f in sorted(files, key=lambda f: f['name'].casefold())
                }

        return {}



"""
Fixed values that do not change during runtime
"""

import os

from xdg.BaseDirectory import xdg_cache_home as XDG_CACHE_HOME
from xdg.BaseDirectory import xdg_config_home as XDG_CONFIG_HOME

from . import __project_name__

DEFAULT_CACHE_DIRECTORY = os.path.join(XDG_CACHE_HOME, __project_name__)
"""Default path cache directory if config option ``cache_directory`` is not set"""

GENERIC_TORRENTS_DIRPATH = os.path.join(DEFAULT_CACHE_DIRECTORY, 'generic_torrents')
"""Path to directory that contains cached torrents for re-using piece hashes"""

CONFIG_FILEPATH = os.path.join(XDG_CONFIG_HOME, __project_name__, 'config.ini')
"""Path to general configuration file"""

TRACKERS_FILEPATH = os.path.join(XDG_CONFIG_HOME, __project_name__, 'trackers.ini')
"""Path to trackers configuration file"""

IMGHOSTS_FILEPATH = os.path.join(XDG_CONFIG_HOME, __project_name__, 'imghosts.ini')
"""Path to image hosting services configuration file"""

CLIENTS_FILEPATH = os.path.join(XDG_CONFIG_HOME, __project_name__, 'clients.ini')
"""Path to BitTorrent clients configuration file"""

VIDEO_FILE_EXTENSIONS = ('mkv', 'mp4', 'avi', 'wmv', 'vob', 'm2ts', 'mts', 'ts')
"""Sequence of file extensions to recognize video files"""

GUESSIT_OPTIONS = {
    'expected_title': [
        'The Collector',
        'xXx',
    ],
    'advanced_config': {
    },
}
"""`options` argument for :meth:`.GuessItApi.guessit`"""



# Copied from upload.php
PTP_TAGS = (
    'action',
    'adventure',
    'animation',
    'arthouse',
    'asian',
    'biography',
    'camp',
    'comedy',
    'crime',
    'cult',
    'documentary',
    'drama',
    'experimental',
    'exploitation',
    'family',
    'fantasy',
    'film.noir',
    'history',
    'horror',
    'martial.arts',
    'musical',
    'mystery',
    'performance',
    'philosophy',
    'politics',
    'romance',
    'sci.fi',
    'short',
    'silent',
    'sport',
    'thriller',
    'video.art',
    'war',
    'western',
)



"""
HTML parsing
"""

import re

from . import LazyModule

bs4 = LazyModule(module='bs4', namespace=globals())


def parse(string):
    """
    Return :class:`~.bs4.BeautifulSoup` instance

    :param string: HTML document

    :raise ContentError: if `string` is invalid HTML
    """
    if isinstance(string, bs4.element.Tag):
        return string
    else:
        return bs4.BeautifulSoup(string, features='html.parser')


def dump(html, filepath):
    """
    Write `html` to `filepath` for debugging

    :param html: String or :class:`~.bs4.BeautifulSoup` instance
    """
    with open(filepath, 'w') as f:
        if isinstance(html, bs4.BeautifulSoup):
            f.write(html.prettify())
        else:
            f.write(parse(str(html)).prettify())


def get(soup, *attributes):
    """
    Get `attributes` from `soup`

    These two calls are equivalent if all attributes exist:

    >>> soup.table.tr.td
    "td value"
    >>> html.get(soup, "table", "tr", "td")
    "td value"

    But if any attribute is `None` (which is what :class:`~.bs4.BeautifulSoup`
    returns for unknown tags), you get `None` instead of forcing you to catch an
    :class:`AttributeError`:

    >>> soup.table.no_such_attribute.td
    AttributeError: 'NoneType' object has no attribute 'td'
    >>> html.get(soup, "table", "no_such_attribute", "td")
    None
    """
    for attr in attributes:
        soup = getattr(soup, attr, None)
        if soup is None:
            return None
    return soup


def as_text(html):
    """Strip HTML tags from string and return text without markup"""
    text = parse(html).get_text()
    text = re.sub(r'(\s)\s+', r'\1', text, flags=re.MULTILINE).strip()
    return text


def purge_javascript(html):
    """
    Return `html` with <script> tags removed

    :param str html: HTML string
    """
    soup = parse(html)
    for script_tag in soup.find_all('script'):
        script_tag.decompose()
    return str(soup)



"""
API for themoviedb.org
"""

import functools
import re

from .. import html, http
from ..types import ReleaseType
from . import common
from .base import WebDbApiBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class TmdbApi(WebDbApiBase):
    """API for themoviedb.org"""

    name = 'tmdb'
    label = 'TMDb'

    default_config = {}

    _url_base = 'http://themoviedb.org'
    _soup_cache = {}

    async def _get_soup(self, path, params={}):
        cache_id = (path, tuple(sorted(params.items())))
        if cache_id in self._soup_cache:
            return self._soup_cache[cache_id]
        text = await http.get(
            url=f'{self._url_base}/{path.lstrip("/")}',
            params=params,
            user_agent=await http.get_popular_user_agent(),
            cache=True,
        )
        self._soup_cache[cache_id] = html.parse(text)
        return self._soup_cache[cache_id]

    async def search(self, query):
        _log.debug('Searching TMDb for %s', query)

        if query.id:
            async def generate_result(id):
                _log.debug('Getting ID: %r', id)
                title_english = await self.title_english(id)
                title_original = await self.title_original(id)
                return _TmdbSearchResult(
                    tmdb_api=self,
                    cast=functools.partial(self.cast, id),
                    directors=functools.partial(self.directors, id),
                    genres=functools.partial(self.genres, id),
                    id=id,
                    summary=functools.partial(self.summary, id),
                    title=title_english or title_original,
                    title_english=functools.partial(self.title_english, id),
                    title_original=functools.partial(self.title_original, id),
                    type=await self.type(id),
                    url=await self.url(id),
                    year=await self.year(id),
                )

            if re.search(r'\b(?:movie|tv)\b', query.id):
                return [await generate_result(query.id)]
            else:
                return [await generate_result(f'movie/{query.id}'),
                        await generate_result(f'tv/{query.id}')]

        elif not query.title:
            return []

        else:
            params = {'query': query.title_normalized}
            if query.year is not None:
                params['query'] += f' y:{query.year}'

            if query.type is ReleaseType.movie:
                soup = await self._get_soup('/search/movie', params=params)
            elif query.type in (ReleaseType.season, ReleaseType.episode):
                soup = await self._get_soup('/search/tv', params=params)
            else:
                movie_results = await self.search(query.copy(type=ReleaseType.movie))
                series_results = await self.search(query.copy(type=ReleaseType.series))
                return movie_results + series_results

            # When search for year, unmatching results are included but hidden.
            for tag in soup.find_all('div', class_='hide'):
                tag.clear()

            items = soup.find_all('div', class_='card')
            results = [_TmdbSearchResult(soup=item, tmdb_api=self)
                       for item in items]

            if query.year is not None:
                # Filter the search results for the queried year because TMDb is
                # very smart and returns wrong search results.
                return [result for result in results
                        if result.year == query.year]
            else:
                return results

    _person_url_path_regex = re.compile(r'(/person/\d+[-a-z]+)')

    def _get_persons(self, tag):
        a_tags = tag.find_all('a', href=self._person_url_path_regex)
        persons = []
        for a_tag in a_tags:
            if a_tag.string:
                name = a_tag.string.strip()
                url_match = self._person_url_path_regex.match(a_tag["href"])
                if url_match:
                    url_path = url_match.group(1)
                    url = f'{self._url_base.rstrip("/")}/{url_path.lstrip("/")}'
                else:
                    url = ''
                persons.append(common.Person(name, url))
        return tuple(persons)

    async def cast(self, id):
        cast = []
        if id:
            soup = await self._get_soup(id)
            cards = soup.select('.people > .card')
            for card in cards:
                cast.extend(self._get_persons(card))
        return tuple(cast)

    async def _countries(self, id):
        return ()

    async def creators(self, id):
        creators = []
        if id:
            soup = await self._get_soup(id)
            profiles = soup.select('.people > .profile')
            for profile in profiles:
                if profile.find('p', string=re.compile(r'(?i:Creator)')):
                    creators.extend(self._get_persons(profile))
        return tuple(creators)

    async def directors(self, id):
        directors = []
        if id:
            soup = await self._get_soup(id)
            profiles = soup.select('.people > .profile')
            for profile in profiles:
                if profile.find('p', string=re.compile(r'(?i:Director)')):
                    directors.extend(self._get_persons(profile))
        return tuple(directors)

    async def genres(self, id):
        genres = ()
        if id:
            soup = await self._get_soup(id)
            genres_tag = soup.find(class_='genres')
            if genres_tag:
                genres = (k.lower() for k in genres_tag.stripped_strings
                          if k != ',')
        return tuple(genres)

    async def poster_url(self, id, season=None):
        if id:
            soup = await self._get_soup(id)
            img_tag = soup.find('img', class_='poster')
            if img_tag:
                srcs = img_tag.get('data-srcset')
                if srcs:
                    path = srcs.split()[0]
                    return f'{self._url_base.rstrip("/")}/{path.lstrip("/")}'
        return ''

    rating_min = 0.0
    rating_max = 100.0

    async def rating(self, id):
        if id:
            soup = await self._get_soup(id)
            rating_tag = soup.find(class_='user_score_chart')
            if rating_tag:
                try:
                    return float(rating_tag['data-percent'])
                except (ValueError, TypeError):
                    pass

    async def _runtimes(self, id):
        runtimes = {}
        if id:
            soup = await self._get_soup(id)
            runtimes_tag = soup.find('span', class_='runtime')
            try:
                text = str(runtimes_tag.string)
            except AttributeError:
                text = ''

            minutes = 0
            for unit, unit_minutes in (('h', 60), ('m', 1)):
                for match in re.finditer(rf'(\d+)\s*{unit}', text):
                    minutes += int(match.group(1)) * unit_minutes
            if minutes > 0:
                runtimes['default'] = minutes

        return runtimes

    _no_overview_texts = (
        "We don't have an overview",
        'No overview found.',
    )

    async def summary(self, id):
        if id:
            soup = await self._get_soup(id)
            overview = ''.join(soup.find('div', class_='overview').stripped_strings)
            if any(text in overview for text in self._no_overview_texts):
                overview = ''
            return overview
        return ''

    async def _title_original(self, id):
        soup = await self._get_soup(id)
        try:
            # Find non-English title
            title_tag = soup.find(string=re.compile(r'Original (?:Title|Name)'))
            parent_tag = title_tag.parent.parent
            strings = tuple(parent_tag.stripped_strings)
            return strings[1]
        except (AttributeError, TypeError, ValueError, IndexError):
            # Default to English title
            english_titles = await self._titles_english(id)
            return english_titles[0]

    async def _titles_english(self, id):
        soup = await self._get_soup(id)
        title_tag = soup.find(class_='title')
        title_parts = list(title_tag.stripped_strings)
        return (title_parts[0],)

    async def type(self, id):
        if id:
            soup = await self._get_soup(id)
            network_tag = soup.find('bdi', string=re.compile(r'^Networks?$'))
            if network_tag:
                return ReleaseType.series
            else:
                return ReleaseType.movie
        else:
            return ReleaseType.unknown

    async def url(self, id):
        if id:
            return f'{self._url_base.rstrip("/")}/{id.strip("/")}'
        return ''

    async def year(self, id):
        if id:
            soup = await self._get_soup(id)
            release_date_tag = soup.find(class_='release_date')
            if release_date_tag:
                year = ''.join(release_date_tag.stripped_strings).strip('()')
                if len(year) == 4 and year.isdigit():
                    return year
        return ''


class _TmdbSearchResult(common.SearchResult):
    def __init__(self, *, tmdb_api, soup=None, cast=None, countries=None,
                 directors=None, id=None, genres=None, summary=None, title=None,
                 title_english=None, title_original=None, type=None, url=None,
                 year=None):
        soup = soup or html.parse('')
        id = id or self._get_id(soup)
        return super().__init__(
            cast=cast or functools.partial(tmdb_api.cast, id),
            countries=countries or (),
            directors=directors or functools.partial(tmdb_api.directors, id),
            id=id,
            genres=genres or functools.partial(tmdb_api.genres, id),
            summary=summary or functools.partial(tmdb_api.summary, id),
            title=title or self._get_title(soup),
            title_english=title_english or functools.partial(tmdb_api.title_english, id),
            title_original=title_original or functools.partial(tmdb_api.title_original, id),
            type=type or self._get_type(soup),
            url=url or self._get_url(soup),
            year=year or self._get_year(soup),
        )

    def _get_id(self, soup):
        a_tag = soup.find('a', class_='result')
        if a_tag:
            href = a_tag.get('href')
            return re.sub(r'^.*/((?:movie|tv)/[0-9]+).*?$', r'\1', href)
        return ''

    def _get_url(self, soup):
        a_tag = soup.find('a', class_='result')
        if a_tag:
            href = a_tag.get('href')
            if href.startswith('http'):
                return href
            elif href.startswith('/'):
                return f'{TmdbApi._url_base}{href}'
            elif href:
                return f'{TmdbApi._url_base}/{href}'
        return ''

    def _get_type(self, soup):
        a_tag = soup.find('a', class_='result')
        if a_tag:
            href = a_tag.get('href')
            if href:
                tmdb_type = re.sub(r'(?:^|/)(\w+)/.*$', r'\1', href)
                if tmdb_type == 'movie':
                    return ReleaseType.movie
                elif tmdb_type == 'tv':
                    return ReleaseType.series
        return ReleaseType.unknown

    def _get_title(self, soup):
        header = soup.select('.result > h2')
        if header:
            return header[0].string
        else:
            return ''

    def _get_year(self, soup):
        release_date = soup.find(class_='release_date')
        if release_date:
            match = re.search(r'(\d{4})$', release_date.string)
            if match:
                return match.group(1)
        return ''



"""
API for querying services like IMDb
"""

from .. import CaseInsensitiveString, subclasses, submodules
from . import imdb, tmdb, tvmaze
from .base import WebDbApiBase
from .common import Person, Query, SearchResult


def webdbs():
    """Return list of :class:`.WebDbApiBase` subclasses"""
    return subclasses(WebDbApiBase, submodules(__package__))


def webdb(name, config=None):
    """
    Create :class:`.WebDbApiBase` instance

    :param str name: Name of the DB. A subclass of :class:`.WebDbApiBase` with
        the same :attr:`~.WebDbApiBase.name` must exist in one of this package's
        submodules.
    :param dict config: User configuration passed to the subclass specified by
        `name`

    :raise ValueError: if no matching subclass can be found

    :return: :class:`.WebDbApiBase` instance
    """
    for cls in webdbs():
        if cls.name == name:
            return cls(config=config)
    raise ValueError(f'Unsupported web database: {name}')


def webdb_names():
    """Return sequence of valid `name` arguments for :func:`.webdb`"""
    return sorted(CaseInsensitiveString(cls.name) for cls in webdbs())



from prompt_toolkit.filters import Condition
from prompt_toolkit.layout.containers import (ConditionalContainer, HSplit,
                                              Window)
from prompt_toolkit.layout.controls import FormattedTextControl

from ....utils import cached_property
from ....utils.types import SceneCheckResult
from .. import widgets
from . import JobWidgetBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class SceneSearchJobWidget(JobWidgetBase):
    def setup(self):
        pass

    @cached_property
    def runtime_widget(self):
        return None


class SceneCheckJobWidget(JobWidgetBase):
    is_interactive = True

    def setup(self):
        self._question = FormattedTextControl('')
        self._radiolist = widgets.RadioList()
        self._activity_indicator = widgets.ActivityIndicator(
            active=True,
            style='class:info',
            extend_width=True,
        )
        self.job.signal.register('finished', lambda _: self._activity_indicator.disable())
        self.job.signal.register('ask_release_name', self._ask_release_name)
        self.job.signal.register('ask_is_scene_release', self._ask_is_scene_release)

    def _ask_release_name(self, release_names):
        _log.debug('Asking for release name: %r', release_names)

        self._activity_indicator.active = False

        def handle_release_name(choice):
            _log.debug('Got release name: %r', choice)
            # Remove current question
            self._radiolist.choices = ()
            self.invalidate()

            # Verify user choice
            self.job.user_selected_release_name(choice[1])
            self._activity_indicator.active = True

        self._radiolist.on_accepted = handle_release_name
        self._radiolist.choices = [(release_name, release_name) for release_name in release_names]
        self._radiolist.choices.append(('This is not a scene release', None))

        self._question.text = 'Pick the correct release name:'
        self.invalidate()

    def _ask_is_scene_release(self, is_scene_release):
        _log.debug('Confirming guess: %r', is_scene_release)

        self._activity_indicator.active = False

        def make_choice(choice):
            if choice:
                string = 'Yes'
                if is_scene_release is SceneCheckResult.true:
                    string += ' (autodetected)'
                return (string, SceneCheckResult.true)
            else:
                string = 'No'
                if is_scene_release is SceneCheckResult.false:
                    string += ' (autodetected)'
                return (string, SceneCheckResult.false)

        def handle_is_scene_release(choice):
            _log.debug('Got decision from user: %r', choice)
            self.job.finalize(choice[1])
            self.invalidate()

        self._radiolist.choices = (make_choice(True), make_choice(False))
        self._radiolist.focused_choice = make_choice(is_scene_release)
        self._radiolist.on_accepted = handle_is_scene_release

        self._question.text = 'Is this a scene release?'
        self.invalidate()

    @cached_property
    def runtime_widget(self):
        return HSplit(
            children=[
                ConditionalContainer(
                    filter=Condition(lambda: self._activity_indicator.active),
                    content=self._activity_indicator,
                ),
                ConditionalContainer(
                    filter=Condition(lambda: bool(self._radiolist.choices)),
                    content=HSplit(
                        children=[
                            Window(self._question),
                            self._radiolist,
                        ],
                    ),
                ),
            ],
        )



"""
Concrete :class:`~.base.TrackerJobsBase` subclass for BHD
"""

import io
import os

from ... import __homepage__, __project_name__, jobs
from ...utils import as_groups, cached_property, release, string
from ..base import TrackerJobsBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class BhdTrackerJobs(TrackerJobsBase):

    @cached_property
    def jobs_before_upload(self):
        return (
            # Background jobs
            self.create_torrent_job,
            self.mediainfo_job,
            self.screenshots_job,
            self.upload_screenshots_job,

            # Interactive jobs
            self.tmdb_job,
            self.imdb_job,
            self.release_name_job,
            self.category_job,
            self.type_job,
            self.source_job,
            self.description_job,
            self.scene_check_job,
            self.tags_job,
        )

    @property
    def isolated_jobs(self):
        """
        Sequence of attribute names (e.g. "imdb_job") that were singled out by the
        user, e.g. with a CLI argument
        """
        if self.options.get('only_description', False):
            # Only activate description_job and its dependencies
            return (self.description_job, self.screenshots_job, self.upload_screenshots_job)
        elif self.options.get('only_title', False):
            # Only activate release_name_job and its dependencies
            return (self.release_name_job, self.imdb_job)
        else:
            # Activate all jobs in jobs_before/after_upload
            return ()

    @cached_property
    def category_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('category'),
            label='Category',
            condition=self.make_job_condition('category_job'),
            prejobs=(
                self.release_name_job,
            ),
            autodetect=self.autodetect_category,
            choices=(
                ('Movie', '1'),
                ('TV', '2'),
            ),
            **self.common_job_args(),
        )

    _autodetect_category_map = {
        'Movie': lambda release_name: release_name.type is release.ReleaseType.movie,
        'TV': lambda release_name: release_name.type in (release.ReleaseType.season,
                                                         release.ReleaseType.episode)
    }

    def autodetect_category(self, _):
        approved_release_name = self.release_name
        _log.debug('Autodetecting category: Approved release type: %r', approved_release_name.type)
        for label, is_match in self._autodetect_category_map.items():
            if is_match(approved_release_name):
                return label

    @cached_property
    def type_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('type'),
            label='Type',
            condition=self.make_job_condition('type_job'),
            prejobs=(
                self.release_name_job,
            ),
            autodetect_and_finish=self.autodetect_type,
            choices=(
                ('UHD 100', 'UHD 100'),
                ('UHD 66', 'UHD 66'),
                ('UHD 50', 'UHD 50'),
                ('UHD Remux', 'UHD Remux'),
                ('BD 50', 'BD 50'),
                ('BD 25', 'BD 25'),
                ('BD Remux', 'BD Remux'),
                ('2160p', '2160p'),
                ('1080p', '1080p'),
                ('1080i', '1080i'),
                ('720p', '720p'),
                ('576p', '576p'),
                ('540p', '540p'),
                ('DVD 9', 'DVD 9'),
                ('DVD 5', 'DVD 5'),
                ('DVD Remux', 'DVD Remux'),
                ('480p', '480p'),
                ('Other', 'Other'),
            ),
            focused='Other',
            **self.common_job_args(),
        )

    _autodetect_type_map = {
        'DVD 9': lambda release_name: release_name.source == 'DVD9',
        'DVD 5': lambda release_name: release_name.source == 'DVD5',
        'DVD Remux': lambda release_name: release_name.source == 'DVD Remux',
        '2160p': lambda release_name: release_name.resolution == '2160p',
        '1080p': lambda release_name: release_name.resolution == '1080p',
        '1080i': lambda release_name: release_name.resolution == '1080i',
        '720p': lambda release_name: release_name.resolution == '720p',
        '576p': lambda release_name: release_name.resolution == '576p',
        '540p': lambda release_name: release_name.resolution == '540p',
        '480p': lambda release_name: release_name.resolution == '480p',
    }

    def autodetect_type(self, _):
        approved_release_name = self.release_name
        _log.debug('Autodetecting type: Approved resolution: %r', approved_release_name.resolution)
        _log.debug('Autodetecting type: Approved source: %r', approved_release_name.source)
        for label, is_match in self._autodetect_type_map.items():
            if is_match(approved_release_name):
                return label

    @cached_property
    def source_job(self):
        # 'output' signal is only emitted when job succeeds while 'finished'
        # signal is also emitted when job fails (e.g. when Ctrl-c is pressed)
        self.release_name_job.signal.register('output', self.autodetect_source)

        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('source'),
            label='Source',
            condition=self.make_job_condition('source_job'),
            prejobs=(
                self.release_name_job,
            ),
            autodetect_and_finish=self.autodetect_source,
            choices=(
                ('Blu-ray', 'Blu-ray'),
                ('HD-DVD', 'HD-DVD'),
                ('WEB', 'WEB'),
                ('HDTV', 'HDTV'),
                ('DVD', 'DVD'),
            ),
            **self.common_job_args(),
        )

    # Map type_job labels to matchers
    _autodetect_source_map = {
        'Blu-ray': lambda release_name: 'BluRay' in release_name.source,
        'HD-DVD': lambda release_name: 'HD-DVD' in release_name.source,
        'WEB': lambda release_name: 'WEB' in release_name.source,
        'HDTV': lambda release_name: 'HDTV' in release_name.source,
        'DVD': lambda release_name: 'DVD' in release_name.source,
    }

    def autodetect_source(self, _):
        approved_release_name = self.release_name
        _log.debug('Autodetecting source: Approved source: %r', approved_release_name.source)
        for label, is_match in self._autodetect_source_map.items():
            if is_match(approved_release_name):
                return label

    @cached_property
    def description_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('description'),
            label='Description',
            condition=self.make_job_condition('description_job'),
            text=self.generate_description,
            autofinish=True,
            read_only=True,
            **self.common_job_args(ignore_cache=True),
        )

    image_host_config = {
        'common': {'thumb_width': 350},
    }

    async def generate_description(self):
        # Wait until all screenshots are uploaded
        await self.upload_screenshots_job.wait()

        rows = []
        screenshot_groups = as_groups(
            self.upload_screenshots_job.uploaded_images,
            group_sizes=(2,),
            default=None,
        )
        for screenshots in screenshot_groups:
            cells = []
            for screenshot in screenshots:
                if screenshot is not None:
                    if screenshot.thumbnail_url is None:
                        raise RuntimeError(f'No thumbnail for {screenshot}')
                    cells.append(f'[url={screenshot}][img]{screenshot.thumbnail_url}[/img][/url]')

            # Space between columns
            rows.append('   '.join(cells))

        screenshots = '[center]\n' + '\n\n'.join(rows) + '\n[/center]'
        promotion = (
            '[align=right][size=1]'
            f'Shared with [url={__homepage__}]{__project_name__}[/url]'
            '[/size][/align]'
        )
        return screenshots + '\n\n' + promotion

    @cached_property
    def tags_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('tags'),
            label='Tags',
            condition=self.make_job_condition('tags_job'),
            text=self.generate_tags,
            autofinish=True,
            read_only=True,
            **self.common_job_args(),
        )

    async def generate_tags(self):
        await self.release_name_job.wait()
        await self.scene_check_job.wait()

        # Any additional tags separated by comma(s). (Commentary, 2in1, Hybrid,
        # OpenMatte, 2D3D, WEBRip, WEBDL, 3D, 4kRemaster, DualAudio, EnglishDub,
        # Personal, Scene, DigitalExtras, Extras)
        tags = []
        if 'WEBRip' in self.release_name.source:
            tags.append('WEBRip')
        elif 'WEB-DL' in self.release_name.source:
            tags.append('WEBDL')
        if 'Hybrid' in self.release_name.source:
            tags.append('Hybrid')
        if self.release_name.has_commentary:
            tags.append('Commentary')
        if self.release_name.has_dual_audio:
            tags.append('DualAudio')
        if 'Open Matte' in self.release_name.edition:
            tags.append('OpenMatte')
        if '2in1' in self.release_name.edition:
            tags.append('2in1')
        if '4k Remastered' in self.release_name.edition:
            tags.append('4kRemaster')
        if self.get_job_attribute(self.scene_check_job, 'is_scene_release'):
            tags.append('Scene')
        if self.options['personal_rip']:
            tags.append('Personal')

        # TODO: 2D3D
        # TODO: 3D
        # TODO: EnglishDub
        # TODO: DigitalExtras
        # TODO: Extras

        return '\n'.join(tags)

    @property
    def post_data(self):
        return {
            'name': self.get_job_output(self.release_name_job, slice=0),
            'category_id': self.get_job_attribute(self.category_job, 'choice'),
            'type': self.get_job_attribute(self.type_job, 'choice'),
            'source': self.get_job_attribute(self.source_job, 'choice'),
            'imdb_id': self.get_job_output(self.imdb_job, slice=0),
            'tmdb_id': self.get_job_output(self.tmdb_job, slice=0).split('/')[1],
            'description': self.get_job_output(self.description_job, slice=0),
            'edition': self.post_data_edition,
            'custom_edition': self.options['custom_edition'],
            'tags': ','.join(self.get_job_output(self.tags_job, slice=0).split('\n')),
            'nfo': self.post_data_nfo,
            'pack': self.post_data_pack,
            'sd': self.post_data_sd,
            'special': self.post_data_special,
            'anon': '1' if self.options['anonymous'] else '0',
            'live': '0' if self.options['draft'] else '1',
        }

    @cached_property
    def post_data_edition(self):
        # The edition of the uploaded release. (Collector, Director, Extended,
        # Limited, Special, Theatrical, Uncut or Unrated)
        edition = self.release_name.edition
        _log.debug('Approved edition: %r', edition)
        if "Collector's Edition" in edition:
            return 'Collector'
        elif "Director's Cut" in edition:
            return 'Director'
        elif 'Extended Cut' in edition:
            return 'Extended'
        elif 'Limited' in edition:
            return 'Limited'
        elif 'Special Edition' in edition:
            return 'Special'
        elif 'Theatrical Cut' in edition:
            return 'Theatrical'
        elif 'Uncut' in edition or 'Uncensored' in edition:
            return 'Uncut'
        elif 'Unrated' in edition:
            return 'Unrated'

    @property
    def post_data_pack(self):
        # The TV pack flag for when the torrent contains a complete season.
        # (0 = No TV pack or 1 = TV Pack). Default is 0
        if self.release_name.type is release.ReleaseType.season:
            return '1'
        else:
            return '0'

    @property
    def post_data_sd(self):
        # The SD flag. (0 = Not Standard Definition, 1 = Standard Definition).
        # Default is 0
        try:
            height = int(self.release_name.resolution[:-1])
        except ValueError:
            return '0'
        else:
            return '1' if height < 720 else '0'

    max_nfo_size = 500_000

    @property
    def post_data_nfo(self):
        # The NFO of the torrent as string.
        if os.path.isdir(self.content_path):
            for entry in os.listdir(self.content_path):
                if entry.lower().endswith('.nfo'):
                    nfo_path = os.path.join(self.content_path, entry)
                    # Limit size to 500kB
                    if os.path.getsize(nfo_path) <= self.max_nfo_size:
                        try:
                            with open(nfo_path, 'rb') as f:
                                return string.autodecode(f.read())
                        except OSError as e:
                            self.error(e.strerror if e.strerror else str(e))

    @property
    def post_data_special(self):
        # The TV special flag for when the torrent contains a TV special. (0 =
        # Not a TV special, 1 = TV Special). Default is 0
        if self.release_name.type is release.ReleaseType.episode:
            if self.options['special']:
                return '1'
        return '0'

    # TODO
    # @property
    # def post_data_region(self):
    #     # The region in which the disc was released. Only for discs! (AUS,
    #     # CAN, CEE, CHN, ESP, EUR, FRA, GBR, GER, HKG, ITA, JPN, KOR, NOR,
    #     # NLD, RUS, TWN or USA)

    @property
    def torrent_filepath(self):
        return self.get_job_output(self.create_torrent_job, slice=0)

    @property
    def mediainfo_filehandle(self):
        mediainfo = self.get_job_output(self.mediainfo_job, slice=0)
        return io.BytesIO(bytes(mediainfo, 'utf-8'))



from prompt_toolkit.filters import Condition
from prompt_toolkit.layout.containers import ConditionalContainer, HSplit

from ....utils import cached_property, fs
from .. import widgets
from . import JobWidgetBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class ScreenshotsJobWidget(JobWidgetBase):
    def setup(self):
        content_name = fs.basename(self.job.kwargs['content_path'])
        self._activity_indicator = widgets.ActivityIndicator(
            format=f'{{indicator}} Analyzing {content_name}',
            active=True,
            style='class:info',
            extend_width=True,
        )
        self._screenshot_progress = widgets.ProgressBar()
        self.job.signal.register('output', self.handle_screenshot_path)
        self.job.signal.register('finished', lambda _: self._activity_indicator.disable())

    def handle_screenshot_path(self, path):
        self._activity_indicator.active = False
        self.invalidate()

        if self.job.screenshots_total > 0:
            self._screenshot_progress.percent = self.job.screenshots_created / self.job.screenshots_total * 100

    @cached_property
    def runtime_widget(self):
        return HSplit(
            children=[
                ConditionalContainer(
                    filter=Condition(lambda: self._activity_indicator.active),
                    content=self._activity_indicator,
                ),
                ConditionalContainer(
                    filter=Condition(lambda: not self._activity_indicator.active),
                    content=self._screenshot_progress,
                ),
            ],
        )



"""
Concrete :class:`~.base.TrackerBase` subclass for PTP
"""

import re
import urllib

from ... import errors, utils
from ..base import TrackerBase
from .config import PtpTrackerConfig
from .jobs import PtpTrackerJobs

import logging  # isort:skip
_log = logging.getLogger(__name__)


class PtpTracker(TrackerBase):
    name = 'ptp'
    label = 'PTP'

    setup_howto_template = (
        'TODO'
    )

    TrackerConfig = PtpTrackerConfig
    TrackerJobs = PtpTrackerJobs

    @property
    def _base_url(self):
        return self.options['base_url']

    @property
    def _ajax_url(self):
        return urllib.parse.urljoin(self._base_url, '/ajax.php')

    @property
    def _logout_url(self):
        return urllib.parse.urljoin(self._base_url, '/logout.php')

    @property
    def _upload_url(self):
        return urllib.parse.urljoin(self._base_url, '/upload.php')

    @property
    def _torrents_url(self):
        return urllib.parse.urljoin(self._base_url, '/torrents.php')

    @property
    def _announce_url(self):
        announce_url = self.options['announce_url']
        if not announce_url:
            raise errors.AnnounceUrlNotSetError(tracker=self)
        else:
            return self.options['announce_url']

    @property
    def _passkey(self):
        # Needed for logging in with ajax.php
        match = re.search(r'.*/([a-zA-Z0-9]+)/announce', self._announce_url)
        if match:
            return match.group(1)
        else:
            raise RuntimeError(f'Failed to find passkey in announce URL: {self._announce_url}')

    async def _request(self, method, *args, error_prefix='', **kwargs):
        # Because HTTP errors (e.g. 404) are raised, we treat RequestErrrors as
        # normal response so we can get the message from the HTML.
        try:
            # `method` is "GET" or "POST"
            response = await getattr(utils.http, method.lower())(
                *args,
                user_agent=True,
                follow_redirects=False,
                **kwargs,
            )
        except errors.RequestError as e:
            response = e

        # Get error from regular exception (e.g. "Connection refused") or the
        # HTML in response.
        try:
            self._maybe_raise_error(response)
        except errors.RequestError as e:
            # Prepend error_prefix to explain the general nature of the error.
            if error_prefix:
                raise errors.RequestError(f'{error_prefix}: {e}')
            else:
                raise e
        else:
            return response

    def _maybe_raise_error(self, response_or_request_error):
        # utils.http.get()/post() raise RequestError on HTTP status codes, but
        # we want to get the error message from the response text.
        # _maybe_raise_error_from_*() handle Response and RequestError.
        self._maybe_raise_error_from_json(response_or_request_error)
        self._maybe_raise_error_from_html(response_or_request_error)

        # If we got a RequestError and we didn't find an error message in the
        # text, we raise it. This handles any real RequestErrors, like
        # "Connection refused". We also raise any other exception so _request()
        # doesn't return it as a regular response.
        if isinstance(response_or_request_error, BaseException):
            raise response_or_request_error

    def _maybe_raise_error_from_json(self, response_or_request_error):
        # Get error message from ajax.php JSON Response or RequestError
        try:
            json = response_or_request_error.json()
        except errors.RequestError:
            # Response or RequestError is not JSON
            pass
        else:
            if (
                isinstance(json, dict)
                and json.get('Result') == 'Error'
                and json.get('Message')
            ):
                raise errors.RequestError(json['Message'])

    def _maybe_raise_error_from_html(self, response_or_request_error):
        # Only attempt to find an error message if this looks like HTML. This
        # prevents a warning from bs4 that tries to be helpful.
        text = str(response_or_request_error)
        if all(c in text for c in '<>\n'):
            doc = utils.html.parse(text)
            try:
                error_header_tag = doc.select('#content .page__title', string=re.compile(r'(?i:error)'))
                error_container_tag = error_header_tag[0].parent
                error_msg_tag = error_container_tag.find('div', attrs={'class': 'panel__body'})
                error_msg = error_msg_tag.get_text().strip()
                if error_msg:
                    raise errors.RequestError(error_msg)
            except (AttributeError, IndexError):
                # No error message found
                pass

    def _get_anti_csrf_token(self, response):
        json = response.json()
        return json['AntiCsrfToken']

    _auth_regex = re.compile(r'logout\.php\?.*\bauth=([0-9a-fA-F]+)')

    async def _get_auth(self):
        assert self.is_logged_in
        response = await self._request('GET', self._base_url)
        doc = utils.html.parse(response)

        try:
            logout_link_tag = doc.find('a', href=self._auth_regex)
            logout_link_href = logout_link_tag['href']
            match = self._auth_regex.search(logout_link_href)
            return match.group(1)
        except (AttributeError, KeyError, TypeError):
            pass

        raise RuntimeError('Could not find auth')

    async def login(self):
        if not self.is_logged_in:
            if not self.options.get('username'):
                raise errors.RequestError('Login failed: No username configured')
            elif not self.options.get('password'):
                raise errors.RequestError('Login failed: No password configured')

            _log.debug('%s: Logging in as %r', self.name, self.options['username'])
            post_data = {
                'username': self.options['username'],
                'password': self.options['password'],
                'passkey': self._passkey,
                # 'keeplogged': '1',
            }
            response = await self._request(
                method='POST',
                url=f'{self._ajax_url}?action=login',
                data=post_data,
                error_prefix='Login failed',
            )
            self._anti_csrf_token = self._get_anti_csrf_token(response)

    async def logout(self):
        try:
            if self.is_logged_in:
                _log.debug('%s: Logging out', self.name)
                await self._request(
                    method='GET',
                    url=self._logout_url,
                    params={'auth': await self._get_auth()},
                    error_prefix='Logout failed',
                )
        finally:
            if hasattr(self, '_anti_csrf_token'):
                delattr(self, '_anti_csrf_token')

    @property
    def is_logged_in(self):
        return bool(getattr(self, '_anti_csrf_token', False))

    async def get_announce_url(self):
        return self._announce_url

    async def upload(self, tracker_jobs):
        post_data = tracker_jobs.post_data
        post_data['AntiCsrfToken'] = self._anti_csrf_token
        _log.debug('Uploading POST data:')
        for k, v in post_data.items():
            _log.debug(' * %s = %r', k, v)

    def normalize_imdb_id(self, imdb_id):
        """
        Format IMDb ID for PTP

        PTP expects 7-characters, right-padded with "0" and without the leading
        "tt".

        If `imdb_id` is `None`, return "0".
        """
        if imdb_id is None:
            return '0'
        else:
            return str(imdb_id).lstrip('t').rjust(7, '0')

    async def get_ptp_group_id_by_imdb_id(self, imdb_id):
        """
        Convert IMDb ID to PTP group ID

        :return: PTP group ID or `None` if PTP doesn't have a group for
            `imdb_id`
        """
        _log.debug('%s: Fetching PTP group ID', imdb_id)
        await self.login()
        response = await self._request(
            method='GET',
            url=self._torrents_url,
            params={
                'imdb': self.normalize_imdb_id(imdb_id),
                'json': '1',
            },
            cache=True,
        )
        match = re.search(r'id=(\d+)', response.headers.get('location', ''))
        if match:
            _log.debug('%s: PTP group ID: %s', imdb_id, match.group(1))
            return match.group(1)
        else:
            _log.debug('%s: No PTP group ID', imdb_id)

    async def get_ptp_metadata(self, imdb_id, key=None):
        """
        Get metadata from PTP website

        :param imdb_id: IMDb ID (e.g. ``tt123456``)
        """
        if key:
            _log.debug('%s: Fetching %s from PTP', imdb_id, key)
        else:
            _log.debug('%s: Fetching metadata from PTP', imdb_id, key)

        assert imdb_id, imdb_id
        await self.login()
        response = await self._request(
            method='GET',
            url=self._ajax_url,
            params={
                'action': 'torrent_info',
                'imdb': self.normalize_imdb_id(imdb_id),
                'fast': '1',
            },
            cache=True,
        )
        results = response.json()
        assert len(results) == 1
        if key is None:
            _log.debug('%s: PTP metadata: %s', imdb_id, results[0])
            return results[0]
        else:
            _log.debug('%s: PTP %s: %s', imdb_id, key, results[0][key])
            return results[0][key]



"""
Concrete :class:`~.base.TrackerConfigBase` subclass for NBL
"""

import base64

from ...utils import configfiles
from .. import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


class NblTrackerConfig(base.TrackerConfigBase):
    defaults = {
        'upload_url': base64.b64decode('aHR0cHM6Ly9uZWJ1bGFuY2UuaW8vdXBsb2FkLnBocA==').decode('ascii'),
        'announce_url': configfiles.config_value(
            value='',
            description=(
                'The complete announce URL with your private passkey.\n'
                'Get it from the website: Shows -> Upload -> Your personal announce URL'
            ),
        ),
        'apikey': configfiles.config_value(
            value='',
            description=(
                'Your personal private API key.\n'
                'Get it from the website: <USERNAME> -> Settings -> API keys'
            ),
        ),
        'source': 'NBL',
        'exclude': [],
    }

    argument_definitions = {
        # Custom arguments, e.g. "upsies submit nbl --foo bar"
    }



"""
Exception classes

Abstraction layers should raise one of these exceptions if an error message
should be displayed to the user. If the programmer made a mistake, any
appropriate builtin exception (e.g. :class:`ValueError`, :class:`TypeError`) or
:class:`RuntimeError` should be raised.

For example, :func:`~.utils.http.get` always raises :class:`RequestError`,
regardless of which library is used or what went wrong, except when something
like caching fails, which is most likely due to a bug.
"""

import json

_NO_DEFAULT_VALUE = object()


class UpsiesError(Exception):
    """Base class for all exceptions raised by upsies"""

    def __eq__(self, other):
        if isinstance(other, type(self)) and str(other) == str(self):
            return True
        else:
            return NotImplemented


class ConfigError(UpsiesError):
    """Error while reading/writing config file or setting config file option"""


class UiError(UpsiesError):
    """Fatal user interface error that cannot be handled by the UI itself"""


class DependencyError(UpsiesError):
    """Some external tool is missing (e.g. ``mediainfo``)"""


class ContentError(UpsiesError):
    """
    Something is wrong with user-provided content, e.g. no video files in the
    given directory or no permission to read
    """

class ProcessError(UpsiesError):
    """Executing subprocess failed"""


class RequestError(UpsiesError):
    """Network request failed"""
    def __init__(self, msg, url='', headers={}, status_code=None, text=''):
        super().__init__(msg)
        self._url = url
        self._headers = headers
        self._status_code = status_code
        self._text = text

    @property
    def url(self):
        """URL that produced this exception"""
        return self._url

    @property
    def headers(self):
        """HTTP headers from server response or empty `dict`"""
        return self._headers

    @property
    def status_code(self):
        """HTTP status code (e.g. 404) or `None`"""
        return self._status_code

    @property
    def text(self):
        """Response string if available"""
        return self._text

    def json(self, default=_NO_DEFAULT_VALUE):
        """
        Parse the :attr:`text` as JSON

        :param default: Return value if :attr:`text` is not valid JSON

        :raise RequestError: if :attr:`text` is not proper JSON and `default` is
            not given
        """
        try:
            return json.loads(self.text)
        except ValueError as e:
            if default is _NO_DEFAULT_VALUE:
                raise RequestError(f'Malformed JSON: {str(self.text)}: {e}')
            else:
                return default


class ScreenshotError(UpsiesError):
    """Screenshot creation failed"""


class ImageResizeError(UpsiesError):
    """Image resizing failed"""


class TorrentError(UpsiesError):
    """Torrent file creation failed"""


def SubprocessError(exception, original_traceback):
    """Attach `original_traceback` to `exception`"""
    exception.original_traceback = f'Subprocess traceback:\n{original_traceback.strip()}'
    return exception


class SceneError(UpsiesError):
    """Base class for scene-related errors"""

class SceneRenamedError(SceneError):
    """Renamed scene release"""
    def __init__(self, *, original_name, existing_name):
        super().__init__(f'{existing_name} should be named {original_name}')
        self._original_name = original_name
        self._existing_name = existing_name

    @property
    def original_name(self):
        """What the release name should be"""
        return self._original_name

    @property
    def existing_name(self):
        """What the release name is"""
        return self._existing_name

class SceneFileSizeError(SceneError):
    """Scene release file size differs from original release"""
    def __init__(self, filename, *, original_size, existing_size):
        super().__init__(f'{filename} should be {original_size} bytes, not {existing_size}')
        self._filename = filename
        self._original_size = original_size
        self._existing_size = existing_size

    @property
    def filename(self):
        """Name of the file in the scene release"""
        return self._filename

    @property
    def original_size(self):
        """Size of the file in the scene release"""
        return self._original_size

    @property
    def existing_size(self):
        """Size of the file in the local file system"""
        return self._existing_size

class SceneMissingInfoError(UpsiesError):
    """Missing information about a file from a scene release"""
    def __init__(self, file_name):
        super().__init__(f'Missing information: {file_name}')

class SceneAbbreviatedFilenameError(UpsiesError):
    """
    An abbreviated scene file name (e.g. foo-bar.mkv) was provided

    These contain almost no information and should always be provided via a
    properly named parent directory.
    """
    def __init__(self, file_name):
        super().__init__(f'Abbreviated scene file name is verboten: {file_name}')


class FoundDupeError(RequestError):
    """
    Tried to upload files that have already been uploaded

    :param filenames: Sequence of file names that have already been uploaded
    """
    def __init__(self, filenames):

        if not filenames:
            msg = ['Potential duplicate files found']

        else:
            files_count = len(filenames)
            if files_count != 1:
                msg = [f'{files_count} potential duplicate files found']
            else:
                msg = ['1 potential duplicate file found']

            msg.append(':\n - ')
            msg.append('\n - '.join(filenames))

        super().__init__(''.join(msg))


class AnnounceUrlNotSetError(RequestError):
    """Announce URL is required but not provided by the user"""
    def __init__(self, *_, tracker):
        super().__init__('Announce URL is not set')
        self._tracker = tracker

    @property
    def tracker(self):
        """:attr:`~.TrackerBase` instance that raised this exception"""
        return self._tracker



"""
Concrete :class:`~.base.TrackerJobsBase` subclass for MTV
"""

import re

from ... import __homepage__, __project_name__, jobs, utils
from ...utils import cached_property
from ...utils.release import ReleaseType
from ..base import TrackerJobsBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class MtvTrackerJobs(TrackerJobsBase):

    torrent_piece_size_max = 8 * 2**20  # 8 MiB

    @cached_property
    def jobs_before_upload(self):
        return (
            # Background jobs
            self.create_torrent_job,
            self.mediainfo_job,
            self.screenshots_job,
            self.upload_screenshots_job,

            # Interactive jobs
            self.category_job,
            self.imdb_job,
            self.scene_check_job,
            self.title_job,
            self.description_job,
        )

    @property
    def isolated_jobs(self):
        if self.options.get('only_description', False):
            return (
                self.mediainfo_job,
                self.screenshots_job,
                self.upload_screenshots_job,
                self.description_job,
            )
        elif self.options.get('only_title', False):
            return (
                self.category_job,
                self.imdb_job,
                self.scene_check_job,
                self.title_job,
            )
        else:
            # Activate all jobs
            return ()

    @cached_property
    def category_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('category'),
            label='Category',
            condition=self.make_job_condition('category_job'),
            autodetect=self.autodetect_category,
            choices=[
                (c['label'], c['value'])
                for c in self._categories
            ],
            **self.common_job_args(),
        )

    _categories = (
        {'label': 'HD Season', 'value': '5', 'type': ReleaseType.season},
        {'label': 'HD Episode', 'value': '3', 'type': ReleaseType.episode},
        {'label': 'HD Movie', 'value': '1', 'type': ReleaseType.movie},
        {'label': 'SD Season', 'value': '6', 'type': ReleaseType.season},
        {'label': 'SD Episode', 'value': '4', 'type': ReleaseType.episode},
        {'label': 'SD Movie', 'value': '2', 'type': ReleaseType.movie},
    )

    _category_value_type_map = {
        c['value']: c['type']
        for c in _categories
    }

    def autodetect_category(self, _):
        # "HD" or "SD"
        if (
            utils.video.height(self.content_path) >= 720
            or utils.video.width(self.content_path) >= 1280
        ):
            resolution = 'HD'
        else:
            resolution = 'SD'

        # "Movie", "Episode" or "Season"
        if self.release_name.type is ReleaseType.movie:
            typ = 'Movie'
        elif self.release_name.type is ReleaseType.season:
            typ = 'Season'
        elif self.release_name.type is ReleaseType.episode:
            typ = 'Episode'
        else:
            raise RuntimeError(f'Unsupported type: {self.release_name.type}')

        category = f'{resolution} {typ}'
        _log.debug('Autodetected category: %r', category)
        return category

    @property
    def chosen_release_type(self):
        """
        :class:`~.types.ReleaseType` enum derived from :attr:`category_job` or
        `None` if :attr:`category_job` is not finished yet
        """
        if self.category_job.is_finished:
            choice = self.get_job_attribute(self.category_job, 'choice')
            return self._category_value_type_map.get(choice)

    release_name_separator = '.'

    release_name_translation = {
        'group': {
            re.compile(r'^NOGROUP$'): r'NOGRP',
        },
    }

    @cached_property
    def title_job(self):
        """
        :class:`~.jobs.dialog.TextFieldJob` instance with text set to the
        release title

        Unlike :attr:`~.TrackerJobsBase.release_name_job`, this uses the
        original scene release name for movie and episode scene releases.
        """
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('title'),
            label='Title',
            prejobs=(self.category_job, self.scene_check_job, self.imdb_job),
            text=self.generate_title,
            condition=self.make_job_condition('title_job'),
            **self.common_job_args(),
        )

    async def generate_title(self):
        assert self.scene_check_job.is_finished
        assert self.imdb_job.is_finished

        if self.chosen_release_type in (ReleaseType.movie, ReleaseType.episode):
            if self.scene_check_job.is_scene_release:
                # Use the scene release name. We rely on scene_check_job making sure
                # that the file/directory name is the correct release name.
                search_results = await utils.predbs.MultiPredbApi().search(self.content_path)
                assert len(search_results) >= 1, search_results
                return search_results[0]

        # Non-scene release or season pack. We may generate a title.
        await self.release_name.fetch_info(webdb=self.imdb, webdb_id=self.imdb_id)
        return str(self.release_name)

    @cached_property
    def description_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('description'),
            label='Description',
            prejobs=(self.mediainfo_job, self.upload_screenshots_job),
            text=self.generate_description,
            condition=self.make_job_condition('description_job'),
            autofinish=True,
            read_only=True,
            hidden=True,
            # Don't cache job output because the number of screenshots can be
            # changed by the user between runs.
            **self.common_job_args(ignore_cache=True),
        )

    async def generate_description(self):
        assert self.mediainfo_job.is_finished
        assert self.upload_screenshots_job.is_finished

        mediainfo = (
            '[mediainfo]'
            + self.get_job_output(self.mediainfo_job, slice=0)
            + '[/mediainfo]'
        )

        screenshots = []
        for screenshot in self.upload_screenshots_job.uploaded_images:
            screenshots.append(f'[img]{screenshot}[/img]')

        screenshots_wrapped = (
            '[spoiler=Screenshots][center]'
            + '\n'.join(screenshots)
            + '[/center][/spoiler]'
        )

        promotion = (
            '[align=right][size=1]'
            f'Shared with [url={__homepage__}]{__project_name__}[/url]'
            '[/size][/align]'
        )

        return (
            mediainfo
            + '\n\n'
            + screenshots_wrapped
            + '\n\n'
            + promotion
        )

    @property
    def post_data_autofill(self):
        return {
            'submit': 'true',
            'MAX_FILE_SIZE': '2097152',
            'fillonly': 'auto fill',
            'category': '0',
            'Resolution': '0',
            'source': '12',
            'origin': '6',
            'title': '',
            'genre_tags': '---',
            'taglist': '',
            'autocomplete_toggle': 'on',
            'image': '',
            'fontfont': '-1',
            'fontsize': '-1',
            'desc': '',
            'fontfont': '-1',
            'fontsize': '-1',
            'groupDesc': '',
            'anonymous': '0',
        }

    @property
    def post_data_upload(self):
        return {
            'submit': 'true',
            'category': self.get_job_attribute(self.category_job, 'choice'),
            'Resolution': '0',
            'source': '12',
            'origin': '6',
            'title': self.get_job_output(self.title_job, slice=0),
            'genre_tags': '---',
            'autocomplete_toggle': 'on',
            'image': '',
            'fontfont': '-1',
            'fontsize': '-1',
            'desc': self.get_job_output(self.description_job, slice=0),
            'fontfont': '-1',
            'fontsize': '-1',
            'groupDesc': '',
            'anonymous': '1' if self.options['anonymous'] else '0',
            'ignoredupes': '1' if self.options['ignore_dupes'] else None,
            'imdbID': self.get_job_output(self.imdb_job, slice=0),
            # 'tmdbID': ...,
            # 'thetvdbID': ...,
            # 'tvmazeID': ...,
        }

    @property
    def torrent_filepath(self):
        return self.get_job_output(self.create_torrent_job, slice=0)



"""
Concrete :class:`~.base.TrackerConfigBase` subclass for PTP
"""

import base64

from ...utils import argtypes, configfiles, imghosts, types
from ..base import TrackerConfigBase


class PtpTrackerConfig(TrackerConfigBase):
    defaults = {
        'base_url': base64.b64decode('aHR0cHM6Ly9wYXNzdGhlcG9wY29ybi5tZQ==').decode('ascii'),
        'username': '',
        'password': '',
        'announce_url': configfiles.config_value(
            value='',
            description='Your personal announce URL',
        ),
        'source': 'PTP',
        'image_host': types.Choice('ptpimg', options=(imghost.name for imghost in imghosts.imghosts())),
        'screenshots': configfiles.config_value(
            value=types.Integer(3, min=3, max=10),
            description='How many screenshots to make',
        ),
        'exclude': [
            r'\.(?i:nfo|txt|jpg|jpeg|png|sfv|md5)$',
            r'/(?i:sample|extra|bonus|feature)',
            r'(?i:sample\.[a-z]+)$',
        ],
    }

    argument_definitions = {
        'submit': {
            ('--screenshots', '--ss'): {
                'help': ('How many screenshots to make '
                         f'(min={defaults["screenshots"].min}, '
                         f'max={defaults["screenshots"].max})'),
                'type': argtypes.number_of_screenshots(
                    min=defaults['screenshots'].min,
                    max=defaults['screenshots'].max,
                ),
            },
            ('--personal-rip', '--pr'): {
                'help': 'Tag as your own encode',
                'action': 'store_true',
            },
            ('--not-main-movie', '--nmm'): {
                'help': 'Upload ONLY contains extras, Rifftrax, Workprints',
                'action': 'store_true',
            },
        },
    }



"""
Concrete :class:`~.base.TrackerBase` subclass for BHD
"""

import re

from ... import errors, utils
from ..base import TrackerBase
from .config import BhdTrackerConfig
from .jobs import BhdTrackerJobs

import logging  # isort:skip
_log = logging.getLogger(__name__)


class BhdTracker(TrackerBase):
    name = 'bhd'
    label = 'BHD'

    setup_howto_template = (
        '{howto.introduction}\n'
        '\n'
        '{howto.current_section}. Announce Passkey\n'
        '\n'
        '   {howto.current_section}.1 On the website, go to My Security -> Passkey and copy\n'
        '       your personal PASSKEY.\n'
        '   {howto.current_section}.2 $ upsies set trackers.bhd.announce_passkey PASSKEY\n'
        '{howto.bump_section}'
        '\n'
        '{howto.current_section}. API key\n'
        '\n'
        '   {howto.current_section}.1 On the website, go to My Security -> API Key and copy\n'
        '       your personal API_KEY.\n'
        '   {howto.current_section}.2 $ upsies set trackers.bhd.apikey API_KEY\n'
        '{howto.bump_section}'
        '\n'
        '{howto.screenshots}\n'
        '\n'
        '{howto.autoseed}\n'
        '\n'
        '{howto.upload}\n'
    )

    TrackerConfig = BhdTrackerConfig
    TrackerJobs = BhdTrackerJobs

    async def login(self):
        pass

    @property
    def is_logged_in(self):
        return True

    async def logout(self):
        pass

    async def get_announce_url(self):
        return '/'.join((
            self.options['announce_url'].rstrip('/'),
            self.options['announce_passkey'],
        ))

    def get_upload_url(self):
        return '/'.join((
            self.options['upload_url'].rstrip('/'),
            self.options['apikey'],
        ))

    DRAFT_UPLOADED_MESSAGE = 'Draft uploaded'

    async def upload(self, tracker_jobs):
        _log.debug('Uploading to %r', self.get_upload_url())

        # Remove None and empty strings from post_data but keep any int(0)
        # values. Also ensure all keys and values are strings.
        post_data = {
            str(k): str(v)
            for k, v in tracker_jobs.post_data.items()
            if v not in (None, '')
        }
        _log.debug('POST data: %r', post_data)

        files = {
            'file': {
                'file': tracker_jobs.torrent_filepath,
                'mimetype': 'application/octet-stream',
            },
            'mediainfo': {
                'file': tracker_jobs.mediainfo_filehandle,
                'filename': 'mediainfo',
                'mimetype': 'application/octet-stream',
            },
        }
        _log.debug('Files: %r', files)

        response = await utils.http.post(
            url=self.get_upload_url(),
            cache=False,
            user_agent=True,
            files=files,
            data=post_data,
        )
        _log.debug('Upload response: %r', response)
        json = response.json()
        _log.debug('Upload response: %r', json)
        try:
            if json['status_code'] == 0:
                # Upload response: {
                #     'status_code': 0,
                #     'status_message': "<error message>",
                #     'success': False,
                # }
                raise errors.RequestError(f'Upload failed: {json["status_message"]}')

            elif json['status_code'] == 1:
                # Upload response: {
                #     'status_code': 1,
                #     'status_message': 'Draft has been successfully saved.',
                #     'success': True,
                # }
                self.warn(json['status_message'])
                self.warn('You have to activate your upload manually '
                          'on the website when you are ready to seed.')
                return tracker_jobs.torrent_filepath

            elif json['status_code'] == 2:
                # Upload response: {
                #     'status_code': 2,
                #     'status_message': '<torrent file URL>',
                #     'success': True,
                # }
                torrent_url = json['status_message']
                return self._torrent_page_url_from_download_url(torrent_url)
            else:
                raise RuntimeError(f'Unexpected response: {str(response)!r}')
        except KeyError:
            raise RuntimeError(f'Unexpected response: {str(response)!r}')

    def _torrent_page_url_from_download_url(self, torrent_download_url):
        # Download URL: .../torrent/download/<torrent name>.123456.d34db33f
        # Website URL: .../torrents/<torrent name>.123456
        torrent_page_url = torrent_download_url.replace('/torrent/download/', '/torrents/')
        torrent_page_url = re.sub(r'\.[a-zA-Z0-9]+$', '', torrent_page_url)
        return torrent_page_url



from prompt_toolkit.layout.containers import (DynamicContainer, HSplit, VSplit,
                                              Window)
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.utils import get_cwidth

from ....utils import browser, cached_property, webdbs
from .. import widgets
from . import JobWidgetBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class WebDbSearchJobWidget(JobWidgetBase):
    def setup(self):
        total_width = 80
        left_column_width = 36
        right_column_width = total_width - left_column_width - 1
        self._widgets = {
            # First row
            'id': widgets.TextField(width=15, style='class:dialog.search.info'),
            'query': widgets.InputField(
                style='class:dialog.search.query',
                text=str(self.job.query),
                on_accepted=self.handle_query,
            ),

            # Right column
            'search_results': _SearchResults(width=right_column_width),
            'title_original': widgets.TextField(
                style='class:dialog.search.info',
                width=right_column_width,
                height=1,
            ),
            'title_english': widgets.TextField(
                style='class:dialog.search.info',
                width=right_column_width,
                height=1,
            ),

            # Left column
            'summary': widgets.TextField(
                style='class:dialog.search.info',
                width=left_column_width,
                height=8,
            ),
            'genres': widgets.TextField(
                style='class:dialog.search.info',
                width=left_column_width,
                height=2,
            ),
            'directors': widgets.TextField(
                style='class:dialog.search.info',
                width=left_column_width,
                height=1,
            ),
            'cast': widgets.TextField(
                style='class:dialog.search.info',
                width=left_column_width,
                height=2,
            ),
            'countries': widgets.TextField(
                style='class:dialog.search.info',
                width=left_column_width,
                height=1,
            ),
        }

        self.set_query_text(self.job.query)

        self.job.signal.register('search_results', self.handle_search_results)
        self.job.signal.register('searching_status', self.handle_searching_status)
        self.job.signal.register('info_updating', self.handle_info_updating)
        self.job.signal.register('info_updated', self.handle_info_updated)
        self.job.signal.register('query_updated', self.handle_query_updated)

    def handle_query(self, buffer):
        query_text = self._widgets['query'].text
        if query_text != self._old_query_text:
            self.job.search(webdbs.Query.from_string(query_text))
            self.set_query_text(self.job.query)
        else:
            # The same query was accepted twice without changing it.
            # Select focused search result.
            selected = self._widgets['search_results'].focused_result
            if selected is not None:
                self.job.result_selected(selected)
            else:
                self.job.result_selected(None)
        self.invalidate()

    def handle_searching_status(self, is_searching):
        self._widgets['search_results'].is_searching = is_searching

    def handle_search_results(self, results):
        self._widgets['search_results'].results = results
        self.invalidate()

    def handle_info_updating(self, attr):
        self._widgets[attr].is_loading = True
        self.invalidate()

    def handle_info_updated(self, attr, value):
        self._widgets[attr].is_loading = False
        self._widgets[attr].text = str(value)
        self.invalidate()

    def handle_query_updated(self, query):
        self.set_query_text(query)

    def set_query_text(self, text):
        self._widgets['query'].set_text(str(text))
        self._old_query_text = self._widgets['query'].text

    @cached_property
    def runtime_widget(self):
        w = self._widgets
        layout = [
            VSplit([
                w['query'],
                widgets.hspacer,
                widgets.HLabel(
                    text='ID',
                    content=w['id'],
                    style='class:dialog.search.label',
                ),
            ]),
            widgets.vspacer,
            VSplit([
                HSplit([
                    widgets.VLabel('Results', w['search_results'], style='class:dialog.search.label'),
                    widgets.VLabel('Original Title', w['title_original'], style='class:dialog.search.label'),
                    widgets.VLabel('Also Known As', w['title_english'], style='class:dialog.search.label'),
                ]),
                widgets.hspacer,
                HSplit([
                    widgets.VLabel('Summary', w['summary'], style='class:dialog.search.label'),
                    widgets.VLabel('Genres', w['genres'], style='class:dialog.search.label'),
                    widgets.VLabel('Director', w['directors'], style='class:dialog.search.label'),
                    widgets.VLabel('Cast', w['cast'], style='class:dialog.search.label'),
                    widgets.VLabel('Country', w['countries'], style='class:dialog.search.label'),
                ]),
            ]),
            widgets.vspacer,
        ]

        @self.bind_key('down')
        @self.bind_key('c-n')
        @self.bind_key('tab')
        def _(event):
            prev_result = self._widgets['search_results'].focused_result
            self._widgets['search_results'].focus_next()
            if prev_result != self._widgets['search_results'].focused_result:
                self.job.result_focused(self._widgets['search_results'].focused_result)

        @self.bind_key('up')
        @self.bind_key('c-p')
        @self.bind_key('s-tab')
        def _(event):
            prev_result = self._widgets['search_results'].focused_result
            self._widgets['search_results'].select_previous()
            if prev_result != self._widgets['search_results'].focused_result:
                self.job.result_focused(self._widgets['search_results'].focused_result)

        # Alt-Enter
        @self.bind_key('escape', 'enter')
        def _(event):
            url = self._widgets['search_results'].focused_result.url
            browser.open(url)

        # Wrapping the HSplit in a VSplit limits the width of the first line
        # ("Search" and "... ID" fields) to the width of the search results +
        # summary, etc. This can probably be removed if someone figured out a
        # way to give the search results + summary a dynamic width.
        return VSplit(
            children=[
                HSplit(
                    children=layout,
                    # key_bindings=self._make_keybindings(),
                    style='class:dialog.search',
                ),
            ],
        )


class _SearchResults(DynamicContainer):
    def __init__(self, results=(), width=40):
        self.results = results
        self._year_width = 4
        self._type_width = 6
        self._title_width = width - self._year_width - self._type_width - 2
        self._activity_indicator = widgets.ActivityIndicator()
        super().__init__(
            lambda: Window(
                content=FormattedTextControl(self._get_text_fragments, focusable=False),
                width=width,
                height=14,
                style='class:dialog.search.results',
            )
        )

    @property
    def is_searching(self):
        return self._activity_indicator.active

    @is_searching.setter
    def is_searching(self, value):
        self._activity_indicator.active = bool(value)

    @property
    def results(self):
        return self._results

    @results.setter
    def results(self, results):
        self._results = tuple(results)
        self._focused_index = 0

    @property
    def focused_result(self):
        if self._results:
            return self._results[self._focused_index]
        else:
            return None

    def focus_next(self):
        if self._focused_index < len(self._results) - 1:
            self._focused_index += 1

    def select_previous(self):
        if self._focused_index > 0:
            self._focused_index -= 1

    def select_first(self):
        self._focused_index = 0

    def select_last(self):
        self._focused_index = len(self._results) - 1

    def _get_text_fragments(self):
        if self.is_searching:
            return [('class:dialog.search.results', self._activity_indicator.text)]
        elif not self._results:
            return 'No results'

        frags = []
        for i, result in enumerate(self._results):
            if i == self._focused_index:
                title_style = 'class:dialog.search.results.focused'
                frags.append(('[SetCursorPosition]', ''))
                self._focused_result = result
            else:
                title_style = 'class:dialog.search.results'

            if get_cwidth(result.title) > self._title_width:
                title = result.title[:self._title_width - 1] + '…'
            else:
                title = result.title
            frags.append((title_style, title.ljust(self._title_width)))

            frags.append(('', (
                ' '
                f'{str(result.year or "").rjust(4)}'
                ' '
                f'{str(result.type).rjust(6)}'
            )))

            frags.append(('', '\n'))
        frags.pop()  # Remove last newline
        return frags



"""
Interactive text user interface and job manager
"""

import collections
import types

from prompt_toolkit.application import Application
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import Layout
from prompt_toolkit.layout.containers import HSplit, Window, to_container
from prompt_toolkit.output import create_output

from ... import utils
from . import jobwidgets, style

import logging  # isort:skip
_log = logging.getLogger(__name__)


class TUI:
    def __init__(self):
        # Map JobBase.name to SimpleNamespace with attributes:
        #   job       - JobBase instance
        #   widget    - JobWidgetBase instance
        #   container - Container instance
        self._jobs = collections.defaultdict(lambda: types.SimpleNamespace())
        self._app = self._make_app()
        self._app_terminated = False
        self._exception = None
        utils.get_aioloop().set_exception_handler(self._handle_exception)

    def _handle_exception(self, loop, context):
        exception = context.get('exception')
        if exception:
            _log.debug('Caught unhandled exception: %r', exception)
            if not self._exception:
                self._exception = exception
            self._exit()

    def _make_app(self):
        self._jobs_container = HSplit(
            # FIXME: Layout does not accept an empty list of children, so we add
            #        an empty Window that doesn't display anything.
            #        https://github.com/prompt-toolkit/python-prompt-toolkit/issues/1257
            children=[Window()],
            style='class:default',
        )
        self._layout = Layout(self._jobs_container)

        kb = KeyBindings()

        @kb.add('escape')
        @kb.add('c-g')
        @kb.add('c-q')
        @kb.add('c-c')
        def _(event, self=self):
            if self._app.is_running:
                self._exit()

        app = Application(
            # Write TUI to stderr if stdout is redirected. This is useful for
            # allowing the user to make decisions in the TUI (e.g. selecting an
            # item from search results) while redirecting the final output
            # (e.g. an IMDb ID).
            output=create_output(always_prefer_tty=True),
            layout=self._layout,
            key_bindings=kb,
            style=style.style,
            full_screen=False,
            erase_when_done=False,
            mouse_support=False,
            on_invalidate=self._update_jobs_container,
        )
        # Make escape key work
        app.timeoutlen = 0.1
        app.ttimeoutlen = 0.1
        return app

    def add_jobs(self, *jobs):
        """Add :class:`~.jobs.base.JobBase` instances"""
        for job in jobs:
            self._add_job(job)

        # Add job widgets to the main container widget (no side effects)
        self._update_jobs_container()

        # Register signal callbacks (no side effects)
        self._connect_jobs(*jobs)

        # Start all enabled jobs. This has side effects because it can trigger
        # more calls to _start_jobs() and _update_jobs_container() via
        # "finished" signal when a job is getting its output from cache.
        self._start_jobs(*jobs)

    def _add_job(self, job):
        if job.name in self._jobs:
            if job is not self._jobs[job.name].job:
                raise RuntimeError(f'Conflicting job name: {job.name}')
        else:
            self._jobs[job.name].job = job
            self._jobs[job.name].widget = jobwidgets.JobWidget(job, self._app)
            self._jobs[job.name].container = to_container(self._jobs[job.name].widget)

    # We accept one argument because the on_invalidate callback passes the
    # Application instance
    def _update_jobs_container(self, _=None):
        enabled_jobs = self._enabled_jobs

        # Unfocus currently focused widget
        for jobinfo in enabled_jobs:
            if jobinfo.widget.is_focused:
                jobinfo.widget.focus_lost()
                break

        # List interactive jobs first
        jobs_container = []
        for jobinfo in enabled_jobs:
            if jobinfo.widget.is_interactive and jobinfo.job.is_started:
                jobs_container.append(jobinfo.container)

                # Focus the first unfinished job
                if not jobinfo.job.is_finished:
                    try:
                        self._layout.focus(jobinfo.container)
                    except ValueError:
                        pass
                        # _log.debug('Unfocusable job: %r', jobinfo.job.name)
                    else:
                        jobinfo.widget.focus_gained()

                        # Don't display more than one unfinished interactive job
                        # unless any job has errors, in which case we are
                        # terminating the application and display all jobs.
                        if not any(jobinfo.job.errors for jobinfo in self._jobs.values()):
                            break

        # Add background jobs below interactive jobs so the interactive widgets
        # don't change position when non-interactive widgets change size.
        for jobinfo in enabled_jobs:
            if not jobinfo.widget.is_interactive:
                jobs_container.append(jobinfo.container)

        self._jobs_container.children[:] = jobs_container

    def _connect_jobs(self, *jobs):
        for job in jobs:
            job.signal.register('finished', self._handle_job_finished)

    def _handle_job_finished(self, finished_job):
        # Terminate application if any job finished with non-zero exit code
        if finished_job.is_finished and finished_job.exit_code != 0:
            if not self._app_terminated:
                _log.debug('Terminating application because of failed job: %r', finished_job.name)
            self._exit()

        else:
            enabled_jobs = [jobinfo.job for jobinfo in self._enabled_jobs]

            # Terminate application if all jobs finished
            if all(job.is_finished for job in enabled_jobs):
                _log.debug('All jobs finished')
                self._exit()

            # Update jobs unless its pointless because _exit() was called
            elif not self._app_terminated:
                # Every time a job finishes, other jobs can become enabled due
                # to the dependencies on other jobs or other conditions. We also
                # want to display the next interactive job when an interactive
                # job is done.
                self._start_jobs(*enabled_jobs)
                self._update_jobs_container()

    def _start_jobs(self, *jobs):
        for job in jobs:
            if not job.is_started and job.autostart:
                job.start()

    @property
    def _enabled_jobs(self):
        return tuple(jobinfo for jobinfo in self._jobs.values()
                     if jobinfo.job.is_enabled)

    def run(self, jobs):
        """
        Block while running `jobs`

        :param jobs: Iterable of :class:`~.jobs.base.JobBase` instances

        :raise: Any exception that occured while running jobs

        :return: :attr:`~.JobBase.exit_code` from the first failed job or 0 for
            success
        """
        self.add_jobs(*jobs)

        # Block until _exit() is called
        self._app.run(set_exception_handler=False)

        exception = self._get_exception()
        if exception:
            _log.debug('Application exception: %r', exception)
            raise exception
        else:
            # First non-zero exit_code is the application exit_code
            for jobinfo in self._enabled_jobs:
                _log.debug('Checking exit_code of %r: %r', jobinfo.job.name, jobinfo.job.exit_code)
                if jobinfo.job.exit_code != 0:
                    return jobinfo.job.exit_code
            return 0

    def _exit(self):
        if not self._app_terminated:
            if not self._app.is_running and not self._app.is_done:
                utils.get_aioloop().call_soon(self._exit)
            else:
                def handle_jobs_terminated(task):
                    try:
                        task.result()
                    except BaseException as e:
                        _log.debug('Handling exception from %r', task)
                        self._exception = e
                    finally:
                        self._app.exit()
                        self._update_jobs_container()

                self._app_terminated = True
                task = self._app.create_background_task(self._terminate_jobs())
                task.add_done_callback(handle_jobs_terminated)

    async def _terminate_jobs(self):
        jobs = [jobinfo.job for jobinfo in self._jobs.values()]

        self._finish_jobs(jobs)

        _log.debug('Waiting for finished jobs before exiting: %s', [j.name for j in jobs])
        for job in jobs:
            if job.is_started and not job.is_finished:
                _log.debug('Waiting for %r', job.name)
                await job.wait()
                _log.debug('Done waiting for %r', job.name)

    def _finish_jobs(self, jobs):
        for job in jobs:
            if not job.is_finished:
                _log.debug('Finishing %s', job.name)
                job.finish()

    def _get_exception(self):
        if self._exception:
            # Exception from _handle_exception()
            return self._exception
        else:
            # First exception from jobs
            for jobinfo in self._enabled_jobs:
                if jobinfo.job.raised:
                    _log.debug('Exception from %s: %r', jobinfo.job.name, jobinfo.job.raised)
                    return jobinfo.job.raised



"""
Create screenshots from video file(s)
"""

import os
import queue

from .. import errors
from ..utils import LazyModule, daemon, fs, image, timestamp, video
from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)

DEFAULT_NUMBER_OF_SCREENSHOTS = 2

natsort = LazyModule(module='natsort', namespace=globals())


class ScreenshotsJob(JobBase):
    """
    Create screenshots from video file(s)

    This job adds the following signals to the :attr:`~.JobBase.signal`
    attribute:

        ``screenshots_total``
            Emitted before screenshots are created. Registered callbacks get the
            total number of screenshots as a positional argument.
    """

    name = 'screenshots'
    label = 'Screenshots'
    cache_id = None

    def initialize(self, *, content_path, timestamps=(), count=0, from_all_videos=False):
        """
        Set internal state

        :param str content_path: Path to file or directory
        :param timestamps: Screenshot positions in the video
        :type timestamps: sequence of "[[H+:]M+:]S+" strings or seconds
        :param count: How many screenshots to make

        :param bool from_all_videos: Whether to take `count` screenshots from
            each video file beneath `content_path` or only from the first video

            See :func:`.video.find_videos` for more information.

            If `content_path` is not a directory, enabling this has no effect.

        If `timestamps` and `count` are not given, screenshot positions are
        picked at even intervals. If `count` is larger than the number of
        `timestamps`, more timestamps are added.
        """
        self._content_path = content_path
        self._screenshots_created = 0
        self._screenshots_total = -1
        self._timestamps = timestamps
        self._count = count
        self._from_all_videos = from_all_videos
        self._screenshots_process = None
        self.signal.add('screenshots_total', record=True)

    def execute(self):
        """
        Execute screenshot creation subprocess

        The subprocess also picks video file(s) and timestamps.
        """
        self._screenshots_process = daemon.DaemonProcess(
            name=self.name,
            target=_screenshots_process,
            kwargs={
                'content_path': self._content_path,
                'timestamps': self._timestamps,
                'count': self._count,
                'from_all_videos': self._from_all_videos,
                'output_dir': self.home_directory,
                'overwrite': self.ignore_cache,
            },
            info_callback=self._handle_info,
            error_callback=self._handle_error,
            finished_callback=self.finish,
        )
        self._screenshots_process.start()

    def finish(self):
        """Terminate screenshot creation subprocess and finish"""
        if self._screenshots_process:
            self._screenshots_process.stop()
        super().finish()

    def _handle_info(self, info):
        if not self.is_finished:
            if 'screenshots_total' in info:
                self._screenshots_total = info['screenshots_total']
                self.signal.emit('screenshots_total', self._screenshots_total)

            if 'screenshot_path' in info:
                self._screenshots_created += 1
                self.send(info['screenshot_path'])

    def _handle_error(self, error):
        if isinstance(error, BaseException):
            self.exception(error)
        else:
            self.error(error)

    @property
    def exit_code(self):
        """`0` if all screenshots were made, `1` otherwise, `None` if unfinished"""
        if self.is_finished:
            if self.screenshots_total < 0:
                # Job is finished but _screenshots_process() never sent us
                # timestamps. That means we're either using previously cached
                # output or the job was cancelled while _screenshots_process()
                # was still initializing.
                if self.output:
                    # If we have cached output, assume the cached number of
                    # screenshots is what the user wanted because the output of
                    # unsuccessful jobs is not cached (see
                    # JobBase._write_output_cache()).
                    return 0
                else:
                    return 1
            elif len(self.output) == self.screenshots_total:
                return 0
            else:
                return 1

    @property
    def screenshots_total(self):
        """
        Total number of screenshots to make

        .. note:: This is ``-1`` until the subprocess that creates the
                  screenshots is executed and determined the number of
                  screenshots.
        """
        return self._screenshots_total

    @property
    def screenshots_created(self):
        """Total number of screenshots made so far"""
        return self._screenshots_created


def _screenshots_process(output_queue, input_queue,
                         *, content_path, timestamps, count, from_all_videos,
                         output_dir, overwrite):
    # Find appropriate video file(s) if `content_path` is a directory
    try:
        video_files = video.find_videos(content_path)
    except errors.ContentError as e:
        output_queue.put((daemon.MsgType.error, str(e)))
    else:
        if not from_all_videos:
            # Only create screenshots from first video
            del video_files[1:]

        timestamps_map = _map_timestamps(
            video_files=video_files,
            timestamps=timestamps,
            count=count,
        )

        # Herald how many screenshots we are going to make
        screenshots_total = sum((len(ts) for ts in timestamps_map.values()))
        output_queue.put((daemon.MsgType.info, {'screenshots_total': screenshots_total}))

        try:
            _screenshot_video_files(
                output_queue, input_queue,
                timestamps_map=timestamps_map,
                output_dir=output_dir,
                overwrite=overwrite,
            )
        except SystemExit:
            # _maybe_terminate signals termination by raising SystemExit. This
            # shouldn't cause any issues as long as we actually exit here.
            pass


def _screenshot_video_files(output_queue, input_queue, *, timestamps_map, output_dir, overwrite):
    # Make all screenshots from all video files
    for video_file, timestamps in timestamps_map.items():
        _maybe_terminate(input_queue=input_queue)
        _screenshot_video_file(
            output_queue, input_queue,
            video_file=video_file,
            timestamps=timestamps,
            output_dir=output_dir,
            overwrite=overwrite,
        )


def _screenshot_video_file(output_queue, input_queue, *, video_file, timestamps, output_dir, overwrite):
    # Make all screenshots from one video file
    for ts in timestamps:
        _maybe_terminate(input_queue=input_queue)
        _make_screenshot(
            output_queue,
            video_file=video_file,
            timestamp=ts,
            output_dir=output_dir,
            overwrite=overwrite,
        )


def _make_screenshot(output_queue, *, video_file, timestamp, output_dir, overwrite):
    # Make one screenshot from one video file
    screenshot_file = os.path.join(
        output_dir,
        fs.basename(video_file) + f'.{timestamp}.png',
    )
    try:
        actual_screenshot_file = image.screenshot(
            video_file=video_file,
            screenshot_file=screenshot_file,
            timestamp=timestamp,
            overwrite=overwrite,
        )
    except errors.ScreenshotError as e:
        output_queue.put((daemon.MsgType.error, str(e)))
    else:
        output_queue.put((daemon.MsgType.info, {'screenshot_path': actual_screenshot_file}))


def _map_timestamps(*, video_files, timestamps, count):
    # Map each video_file to a sequence of timestamps
    timestamps_map = {}
    for video_file in video_files:
        timestamps_map[video_file] = _validate_timestamps(
            video_file=video_file,
            timestamps=timestamps,
            count=count,
        )
    return timestamps_map


def _validate_timestamps(*, video_file, timestamps, count):
    # Validate, normalize, deduplicate and sort timestamps

    # Stay clear of the last 10 seconds
    duration = video.duration(video_file) - 10

    timestamps_pretty = set()
    for ts in timestamps:
        ts = max(0, min(duration, timestamp.parse(ts)))
        timestamps_pretty.add(timestamp.pretty(ts))

    if not timestamps and not count:
        count = DEFAULT_NUMBER_OF_SCREENSHOTS

    # Add more timestamps if the user didn't specify enough
    if count > 0 and len(timestamps_pretty) < count:
        # Convert timestamp strings to seconds
        timestamps = sorted(timestamp.parse(ts) for ts in timestamps_pretty)

        # Fractions of video duration
        positions = [ts / duration for ts in timestamps]

        # Include start and end of video
        if 0.0 not in positions:
            positions.insert(0, 0.0)
        if 1.0 not in positions:
            positions.append(1.0)

        # Insert timestamps between the two positions with the largest distance
        while len(timestamps_pretty) < count:
            pairs = [(a, b) for a, b in zip(positions, positions[1:])]
            max_distance, pos1, pos2 = max((b - a, a, b) for a, b in pairs)
            position = ((pos2 - pos1) / 2) + pos1
            timestamps_pretty.add(timestamp.pretty(duration * position))
            positions.append(position)
            positions.sort()

    return natsort.natsorted(timestamps_pretty)


def _maybe_terminate(*, input_queue):
    try:
        typ, msg = input_queue.get_nowait()
    except queue.Empty:
        pass
    else:
        if typ == daemon.MsgType.terminate:
            raise SystemExit('Terminated')



"""
Non-interactive job manager

This should be used instead of :class:`~.tui.TUI` when :func:`~.utils.is_tty`
returns `False`.

Headless usage is limited to :attr:`~.jobs.base.JobBase.hidden` jobs.
"""

import sys

from ... import errors, utils

import logging  # isort:skip
_log = logging.getLogger(__name__)


class Headless:
    def __init__(self):
        self._jobs = {}
        self._exception = None

    def add_jobs(self, *jobs):
        """Add :class:`~.jobs.base.JobBase` instances"""
        for job in jobs:
            if job.name in self._jobs:
                raise RuntimeError(f'Job was already added: {job.name}')
            elif not job.hidden:
                raise errors.UiError('Headless mode is not supported for this command.')
            else:
                self._jobs[job.name] = job

                # Terminate application if all jobs finished
                job.signal.register('finished', self._exit_if_all_jobs_finished)

                # Terminate application if any job finished with non-zero exit code
                job.signal.register('finished', self._exit_if_job_failed)

    def run(self, jobs):
        """
        Block while running `jobs`

        :param jobs: Iterable of :class:`~.jobs.base.JobBase` instances

        :raise: Any exception that occured while running jobs

        :return: :attr:`~.JobBase.exit_code` from the first failed job or 0 for
            success
        """
        self.add_jobs(*jobs)

        for job in self._enabled_jobs:
            _log.debug('Starting job: %r', job.name)
            job.start()
        utils.get_aioloop().run_until_complete(self._wait_for_jobs())

        # First non-zero exit_code is the application exit_code
        for job in self._enabled_jobs:
            _log.debug('Checking exit_code of %r: %r', job.name, job.exit_code)
            if job.exit_code != 0:
                for error in job.errors:
                    print(f'{job.name}: {error}', file=sys.stderr)
                return job.exit_code
        return 0

    @property
    def _enabled_jobs(self):
        return tuple(job for job in self._jobs.values() if job.is_enabled)

    def _exit_if_all_jobs_finished(self, *_):
        if all(job.is_finished for job in self._enabled_jobs):
            _log.debug('All jobs finished')
            self._finish_jobs()

    def _exit_if_job_failed(self, job):
        if job.is_finished and job.exit_code != 0:
            _log.debug('Terminating application because of failed job: %r', job.name)
            self._finish_jobs()

    async def _wait_for_jobs(self):
        _log.debug('Terminating all jobs')
        for job in self._enabled_jobs:
            if job.is_started and not job.is_finished:
                _log.debug('Waiting for %r', job.name)
                await job.wait()
                _log.debug('Done waiting for %r', job.name)

    def _finish_jobs(self):
        for job in self._enabled_jobs:
            if not job.is_finished:
                _log.debug('Finishing %s', job.name)
                job.finish()



"""
Lookup country information
"""

import functools

from countryguess import guess_country as _guess_country


def _list_or_string(func):
    @functools.wraps(func)
    def wrapper(arg):
        if isinstance(arg, str):
            return func(arg)
        else:
            return tuple(func(country) for country in arg)

    return wrapper


@_list_or_string
def name(country):
    """
    Convert fuzzy country name to consistent name

    :param country: Country name or code, e.g. "Russian Federation" (official
        name), "USA" (common abbreviation), "Korea" (short name), "fr"
        (2-character country code), etc.
    :type country: str or sequence

    :return: Country name or `country` if it can't be associated with any
        country
    """
    return _guess_country(country, attribute='name_short', default=country)


@_list_or_string
def iso2(country):
    """
    Convert fuzzy country name to ISO 3166-1 alpha-2 codes

    :param country: Country name or code, e.g. "Russian Federation" (official
        name), "USA" (common abbreviation), "Korea" (short name), "fr"
        (2-character country code), etc.
    :type country: str or sequence

    :return: Country code or `country` if it can't be associated with any
        country
    """
    return _guess_country(country, attribute='iso2', default=country)


@_list_or_string
def tld(country):
    """
    Convert fuzzy country name to top level domain

    :param country: Country name or code, e.g. "Russian Federation" (official
        name), "USA" (common abbreviation), "Korea" (short name), "fr"
        (2-character country code), etc.
    :type country: str or sequence

    :return: Country TLD or `country` if it can't be associated with any country
    """
    def as_tld(tld):
        return ('uk' if tld == 'GB' else tld).lower()

    return as_tld(iso2(country))



"""
Abstract base class for scene release databases
"""

import abc
import asyncio
import collections
import copy

from ... import constants, errors, utils

import logging  # isort:skip
_log = logging.getLogger(__name__)

natsort = utils.LazyModule(module='natsort', namespace=globals())


class PredbApiBase(abc.ABC):
    """Base class for scene release database APIs"""

    def __init__(self, config=None):
        self._config = copy.deepcopy(self.default_config)
        if config is not None:
            self._config.update(config.items())

    @property
    @abc.abstractmethod
    def name(self):
        """Unique name of the scene release database"""

    @property
    @abc.abstractmethod
    def label(self):
        """User-facing name of the scene release database"""

    @property
    def config(self):
        """
        User configuration

        This is a deep copy of :attr:`default_config` that is updated with the
        `config` argument from initialization.
        """
        return self._config

    @property
    @abc.abstractmethod
    def default_config(self):
        """Default user configuration as a dictionary"""

    @abc.abstractmethod
    async def _search(self, keywords, group=None):
        pass

    async def search(self, query, only_existing_releases=True):
        """
        Search for scene release

        If there are no results and `query` is a directory path that looks like
        a season pack, perform one search per video file in that directory or
        any subdirectory. This is necessary to find mixed season packs.

        :param query: :class:`~.SceneQuery` object or :class:`str` to pass to
            :meth:`~.SceneQuery.from_string` or :class:`collections.abc.Mapping`
            to pass to :meth:`~.SceneQuery.from_release`
        :param bool only_existing_releases: If this is truthy, imaginary season
            pack releases are created and added to the search results.

        :return: :class:`list` of release names as :class:`str`

        :raise RequestError: if the search request fails
        """
        path = None
        if isinstance(query, str):
            path = query
            query = utils.predbs.SceneQuery.from_string(query)
        elif isinstance(query, collections.abc.Mapping):
            query = utils.predbs.SceneQuery.from_release(query)

        try:
            results = list(await self._search(query.keywords, group=query.group))
        except NotImplementedError:
            raise errors.RequestError(f'{self.name} does not support this query')

        if not results and path:
            # Maybe `path` points to season pack?
            # Find episodes and search for them individually.
            return await self._search_for_episodes(path, only_existing_releases)
        else:
            return self._postprocess_search_results(results, query, only_existing_releases)

    async def _search_for_episodes(self, path, only_existing_releases):
        combined_results = []
        for episode_query in self._generate_episode_queries(path):
            results = await self.search(episode_query, only_existing_releases)
            combined_results.extend(results)
        return combined_results

    def _generate_episode_queries(self, path):
        info = utils.release.ReleaseInfo(path)
        if info['type'] is utils.release.ReleaseType.season:
            # Create SceneQuery from each episode path
            episode_paths = utils.fs.file_list(path, extensions=constants.VIDEO_FILE_EXTENSIONS)
            for episode_path in episode_paths:
                if not utils.predbs.is_abbreviated_filename(episode_path):
                    _log.debug('Generating query for episode: %r', episode_path)
                    # guessit prefers getting the group name from the parent
                    # directory, but the group in the parent directory is likely
                    # "MiXED", so we definitely want the group from the file.
                    filename = utils.fs.basename(episode_path)
                    yield utils.predbs.SceneQuery.from_string(filename)

    def _postprocess_search_results(self, results, query, only_existing_releases):
        def sorted_and_deduped(results):
            return natsort.natsorted(set(results), key=str.casefold)

        if not query.episodes:
            # _log.debug('No episodes queried: %r', query.episodes)
            return sorted_and_deduped(results)
        else:
            # _log.debug('Episodes queried: %r', query.episodes)

            def get_wanted_episodes(season):
                # Combine episodes from any season with episodes from given
                # season (season being empty string means "any season")
                eps = None
                if '' in query.episodes:
                    eps = query.episodes['']
                if season in query.episodes:
                    eps = (eps or []) + query.episodes[season]
                return eps

            # Translate single episodes into season packs.
            matches = []
            for result in results:
                for result_season, result_eps in utils.release.Episodes.from_string(result).items():
                    wanted_episodes = get_wanted_episodes(result_season)

                    # [] means season pack
                    if wanted_episodes == []:
                        if only_existing_releases:
                            # Add episode from wanted season pack
                            _log.debug('Adding episode from season pack: %r', result)
                            matches.append(result)
                        else:
                            season_pack = utils.predbs.common.get_season_pack_name(result)
                            if season_pack not in matches:
                                _log.debug('Adding season pack: %r', season_pack)
                                matches.append(season_pack)

                    elif wanted_episodes is not None:
                        for ep in result_eps:
                            if ep in wanted_episodes:
                                matches.append(result)
                                break

            return sorted_and_deduped(matches)

    @abc.abstractmethod
    async def _release_files(self, release_name):
        pass

    async def release_files(self, release_name):
        """
        Map release file names to file information

        If this is not implemented by the subclass, :class:`NotImplementedError`
        is raised.

        Each file information is a dictionary that contains at least the keys
        ``release_name``, ``file_name`` and ``size``. More keys may be available
        depending on the subclass implementation.

        If `release_name` is a season pack, information the relevant episode
        releases is returned.

        :param str release_name: Exact name of the release

        :raise RequestError: if request fails or `release_name` is not found
        """
        try:
            files = await self._release_files(release_name)
        except NotImplementedError:
            raise errors.RequestError(f'{self.name} does not support this query')

        if files:
            return files
        else:
            _log.debug('No such release: %r', release_name)
            files = {}

        # If scene released "Foo.S01E0{1,2,3,...}.720p.BluRay-BAR" and we're
        # searching for "Foo.S01.720p.BluRay-BAR", we most likely don't get any
        # results. But we can get release names of individual episodes by
        # searching for the season pack, and then we can call release_files()
        # for each episode.
        release_info = utils.release.ReleaseInfo(release_name)
        if release_info['type'] is utils.release.ReleaseType.season:
            results = await self.search(release_info, only_existing_releases=True)
            if results:
                files = await asyncio.gather(
                    *(self._release_files(result) for result in results)
                )

                # Flatten sequence of dictionaries into single dictionary
                files = {
                    file_name: file_info
                    for files_ in files
                    for file_name, file_info in files_.items()
                }
                _log.debug('Season pack from multiple episode releases: %r', files)

        # If scene released season pack (e.g. Extras or Bonus content) and we're
        # searching for a single episode, we most likely don't get any results.
        # Search for the season pack to get all files.
        elif release_info['type'] is utils.release.ReleaseType.episode:
            # Remove single episodes from seasons
            release_info['episodes'].remove_specific_episodes()
            results = await self.search(release_info)
            if len(results) == 1:
                _log.debug('Getting files from single result: %r', results[0])
                files = await self._release_files(results[0])

        # Go through all files and find the exact release name we're looking for.
        # Don't do this exclusively for episodes because not all multi-file releases
        # are a list of episodes (e.g. extras may not contain any "Exx").
        for file_name, file_info in files.items():
            if utils.fs.strip_extension(release_name) == utils.fs.strip_extension(file_name):
                files = {file_name: file_info}
                _log.debug('Single file from season pack release: %r', files)
                break

        return files
